import os
from string import Template
import shutil
from copy import copy, deepcopy
import json
import contextlib
from io import BytesIO
from collections import Iterable
from pprint import pprint

from PIL import Image, ImageOps
import click
import numpy
from ruamel.yaml import YAML
import regex as re #Backwards compatible!

from tokenimaging import TokenImage
from helpers import *
from packs import *

def loadConfig(onlyUpdateMessage=False):
	"""Load configuration files. Include packs config unless omitted. Also, return the update version message."""
	
	#Load main YAML file.
	yaml = YAML(typ='safe')
	with open('config/mainconfig.yaml') as yamlFile:
		documents = yaml.load_all(yamlFile)
		updateMessage, setsData, formatsData, packsData = list(documents)
	
		if not onlyUpdateMessage:
			loadPacksConfig() #Gets the structures
			
			#Process.
			for key, value in setsData.items():
				Category.fromYAML(key, value)
			for key, value in formatsData.items():
				Format.fromYAML(key, value)
			for key, value in packsData.items():
				PackCategory.fromYAML(key, value)
	
	return updateMessage

def fetchContainer(key):
	try:
		return Category[key]
	except KeyError:
		try:
			return Block[key]
		except KeyError:
			try:
				return Set[key]
			except KeyError:
				raise KeyError(f"No category, block, or set found for key '{key}'.")

#File output

def writeUpdateList():
	with open('../templates/updatelist-template.txt') as f:
		template = Template(f.read())
	
	path = 'plugins/magic/'
	url = 'http://magicplugin.normalitycomics.com/update/'
	rootFiles = ('version', 'formats', 'ListOfCardDataFiles')
	categories = Category.instancesList()
	numberOfPackFiles = PackCategory.numberOfPackFiles
	
	display(2, 'Computing file checksums')
	updateLines = []
	for name in rootFiles:
		f = name + '.txt'
		local = path + f
		remote = url + f
		checksum = str(getChecksum(os.path.join(updateFilesDir, f)))
		updateLines.append('\t'.join((local, remote, checksum)))
	
	for category in categories:
		f = category.name + '.txt'
		local = path + 'sets/' + f
		remote = url + 'cardFiles/' + f
		checksum = str(getChecksum(os.path.join(updateFilesDir, 'cardFiles',  f)))
		updateLines.append('\t'.join((local, remote, checksum)))
	
	for i in range(1, numberOfPackFiles + 1):
		f = f'packdefinitions{i}.xml'
		local = path + 'packs/' + f
		remote = url + 'packs/' + f
		checksum = str(getChecksum(os.path.join(updateFilesDir, 'packs', f)))
		updateLines.append('\t'.join((local, remote, checksum)))
	
	now = datetime.now()
	values = {
		'year': str(now.year)[-2:],
		'month': str(now.month).zfill(2),
		'day': str(now.day).zfill(2),
		'updateLines': '\n'.join(updateLines)
	}
	fileName = os.path.join(updateFilesDir, 'updatelist.txt')
	display(1, f'Writing updatelist file: {fileName}')
	with open(fileName, 'w') as f:
		f.write(template.substitute(values))

def writeVersionFile(message):
	with open('../templates/version-template.txt') as f:
		template = Template(f.read())
	
	now = datetime.now()
	values = {
		'year': str(now.year)[-2:],
		'month': str(now.month).zfill(2),
		'day': str(now.day).zfill(2),
		'versionUrl': versionUrl,
		'updateUrl': updateUrl,
		'message': message,
	}
	fileName = os.path.join(updateFilesDir, 'version.txt')
	display(1, 'Writing version file: ' + fileName)
	display(2, f"\tMesssage: '{message}'")
	os.makedirs(os.path.dirname(fileName), exist_ok=True)
	with open(fileName, 'w') as f:
		f.write(template.substitute(values))

#The nuts and bolts of Magic card lists

class Category(metaclass=Recorded):
	"""
	A category of sets, which will be saved
	as a data file for the plugin
	"""
	def __init__(self, name, contents, defaultDrafted=True):
		self.name = name
		self.contents = list(contents)
		self.defaultDrafted = defaultDrafted
	
	@classmethod
	def fromYAML(cls, categoryKey, categoryValue):
		contents = []
		defaultDrafted=True
		
		for key, value in categoryValue.items():
			if ' Block' in key:
				contents.append(Block.fromYAML(key, value))
			elif key == 'defaultDrafted' and value is False:
				defaultDrafted = False
			else:
				contents.append(Set.fromYAML(key, value))
		
		return cls(categoryKey, contents, defaultDrafted=defaultDrafted)
	
	@property
	def sets(self):
		"""Returns all sets, stripped from blocks"""
		sets = []
		for item in self.contents:
			if hasattr(item, 'sets'):
				sets += list(item.sets)
			else:
				sets.append(item)
		return sets
	
	@property
	def cards(self):
		return [card for set_ in self.sets for card in set_.cards]
	
	@property
	def tokens(self):
		return [token for set_ in self.sets for token in set_.tokens]
	
	@property
	def packs(self):
		return filter(None, [set_.pack for set_ in self.sets])
	
	def __repr__(self):
		return f'<Category: {self.name}>'
	
	def getCards(self):
		"""Tell all contained sets to get their cards."""
		if self.name == 'lackeyccg-custom-tokens':
			return
		display(1, f'Beginning category: {self.name}')
		for set_ in self.sets:
			set_.getCards()
		display(1, '')
		#Remove "notInPacks" from cards if this whole category isn't drafted.
		if not self.defaultDrafted:
			for set_ in self.sets:
				if not set_.isDrafted:
					for card in set_.cards:
						card.draftQualities = ''
	
	def output(self):
		"""How to output this category in the formats file"""
		return f'<filetoinclude>{self.name}.txt</filetoinclude>\n'
	
	def resolveSpawnScripts(self, review=False):
		if self.name == 'lackeyccg-custom-tokens':
			return
		for set_ in self.sets:
			set_.resolveSpawnScripts(review=review)
	
	def writeCardFile(self, destDir=os.path.join(updateFilesDir, 'cardFiles')):
		"""Write this category's card data file"""
		fileName = os.path.join(destDir, f'{self.name}.txt')
		os.makedirs(os.path.dirname(fileName), exist_ok=True)
		display(1, f'Writing {fileName}...')
		headers = [
			'Name',
			'Set',
			'ImageFile',
			'Color',
			'ColorID',
			'Cost',
			'ConvertedCost',
			'Type',
			'Power',
			'Toughness',
			'Loyalty',
			'Rarity',
			'DraftQualities',
			'Sound',
			'Script',
			'Text',
		]
		
		with open(fileName, 'w', encoding='ascii') as f:
			f.write('\t'.join(headers) + '\n\n\n')
			for set_ in self.sets:
				f.write(set_.output())
	
	@classmethod
	def writeListOfCardDataFiles(cls):
		"""Write card data list file"""
		fileName = os.path.join(updateFilesDir, 'ListOfCardDataFiles.txt')
		os.makedirs(os.path.dirname(fileName), exist_ok=True)
		display(1, f'Writing {fileName}...')
		with open(fileName, 'w', encoding='ascii') as f:
			f.write('<listofcarddatafiles>\n\n')
			for category in cls.instancesList():
				f.write(category.output())
			f.write('\n</listofcarddatafiles>\n')

class Block(metaclass=Recorded):
	"""A block"""
	def __init__(self, name, sets):
		self.name = name
		self.sets = list(sets)
		for set_ in self.sets:
			set_.block = self
	
	@classmethod
	def fromYAML(cls, blockKey, blockValue):
		sets = [Set.fromYAML(setKey, setValue) for setKey, setValue in blockValue.items()]
		return cls(blockKey, sets)
	
	def __repr__(self):
		return f'<Block: {self.name}>'
	
	@property
	def cards(self):
		return [card for set_ in self.sets for card in set_.cards]
	
	@property
	def tokens(self):
		return [token for set_ in self.sets for token in set_.tokens]

class Set(metaclass=Recorded):
	"""Represents a Magic set. Contains information
	necessary for generating the Lackey files"""
	#Key sets by code rather than by full name.
	key = 'code'
	
	def __init__(self, code, name, customCode=None, **kwargs):
		"""Construct the set object, but don't query Scryfall for cards yet."""		
		#There are so many optional arguments with default
		#values that this is neater than jamming them all up there.
		#display(2, f'Creating set: {name}')
		self.name = name
		self.code = code
		self.customCode = customCode
		self.temp = kwargs.get('temp', False)
		self.tempNumbers = [str(item) for item in kwargs.get('tempNumbers', [])]
		self.getImages = kwargs.get('getImages', False)
		self.new = kwargs.get('new', None)
		if self.new is None:
			if self.getImages:
				self.new = True
			else:
				self.new = False
		self.customQuery = kwargs.get('customQuery', None)
		self.oversize = kwargs.get('oversize', False)
		self.onlyTokens = kwargs.get('onlyTokens', False)
		self.cards = []
		self.partials = [str(item) for item in kwargs.get('partials', [])]
		self.spawnBlanks = kwargs.get('spawnBlanks', False)
		self.tokensFetched = False #Has this set's token's been fetched?
		#Drafting
		#If isDrafted isn't set, the set creates a booster pack.
		#We just don't use it if it's not defined in pack categories.
		self.isDrafted = kwargs.get('isDrafted', None)
		self.promos = kwargs.get('promos', None)
		self.excludePromos = kwargs.get('excludePromos', False)
		self.draftQualities = kwargs.get('draftQualities')
		if self.draftQualities is not None and isinstance(self.draftQualities, str):
			self.draftQualities = [self.draftQualities]
		self.packStructure = kwargs.get('packStructure', None)
		if self.isDrafted is not False:
			self.pack = Booster(self)
		else:
			self.pack = None
	
	@classmethod
	def fromYAML(cls, setKey, setValue):
		if isinstance(setValue, str):
			return cls(setKey, setValue)
		elif isinstance(setValue, dict):
			return cls(setKey, **setValue)
		else:
			return cls(setKey, *setValue)
	
	@property
	def packs(self):
		"""Provide the pack as a list, to provide same interface as blocks and categories."""
		if self.pack:
			return [self.pack]
		else:
			return []
	
	@indented
	def getCards(self, onlyTokens=False):
		"""
		Query Scryfall, using any optional parameters
		defined for this set, and return a list of card objects.
		(Should it return or assign?)
		
		Should also download images if that's set.
		"""
		if self.code == 'lackey':
			gunk = Card(name='[lackey] Gunk', set_=self, color='', type_='sorcery', cmc='0', rarity='T', text="Gunk is a token. | Cycling {4} | (This isn't a real card. It's only in here for Gunk Slug.")
			self.cards.append(gunk)
			self.tokensFetched = True
			return #Custom token set is different
		
		display(1, f'Getting cards for {self.name}...')
		Indent.increase()
		#Set the query for Scryfall:
		q = 'e:' + self.code
		if self.customQuery:
			q += f' ({self.customQuery})' #Parentheses are important! Makes sure a top-level "and" join
		
		params = {
			'order': 'set',
			'q': q,
			'unique': 'cards',
		}
		
		#Get all the variants for Unstable. Hardcoded for now, since it's the only set where
		#I want to do that.
		if self.code == 'ust':
			params['unique'] = 'prints'
		
		jsonCards = []
		promoJsonCards = []
		
		#If this set is downloading images, don't use the cache.
		cached = not self.new
		
		if not self.onlyTokens and not onlyTokens: #first is from the set's setting, second is from this method call
			if self.oversize != 'only':
				cardParams = params.copy()
				#Kludgey hardcoding for now to make Wastes work. Probably shouldn't be located here:
				if self.code == 'ogw':
					cardParams['unique'] = 'art'
				display(2, f'Requesting cards (Using cache: {cached})...')
				jsonCards += scryfallRequest(cardParams, cached=cached)
		
			if self.oversize:
				#Get oversize cards (like schemes and planes)
				display(2, f'Requesting oversize (Using cache: {cached})...')
				oversizeParams = params.copy()
				oversizeParams['q'] = oversizeParams['q'].replace('e:', 'e:o')
				jsonCards += scryfallRequest(oversizeParams, cached=cached)
		
		#Get tokens too, if any.
		display(2, f'Requesting tokens (Using cache: {cached})...')
		tokenParams = params.copy()
		tokenParams['q'] = tokenParams['q'].replace('e:', 'e:t')
		jsonCards += scryfallRequest(tokenParams, cached=cached)
		self.tokensFetched = True
		
		#Get promos, if requested.
		if self.promos:
			display(2, 'Requesting promos...')
			promoJsonCards = {}
			
			Indent.increase()
			for code, customQuery in self.promos.items():
				display(2, f'Requesting {code}...')
				promoJsonCards[code] = []
				promoParams = params.copy()
				customQuery = cnRange(customQuery, self.code, cached=cached) #Convert query if it contains comparison operators for collector number.
				promoParams['q'] = q + f'({customQuery})'
				promoJsonCards[code] += scryfallRequest(promoParams, cached=cached)
			Indent.decrease()
			
		if not jsonCards and not onlyTokens:
			raise LookupError(f'No cards were returned for {self.name}')
		
		display(2, 'Parsing card list...')
		cards = []
		for jsonCard in jsonCards:
			#Remember, this always returns a LIST,
			#even if it's just a one-sided card.
			cards += Card.fromRawScryfall(self, jsonCard)
		
		if promoJsonCards:
			display(2, 'Parsing promo cards...')
			Indent.increase()
			for code, thisPromoJsonCards in promoJsonCards.items():
				display(2, f'Parsing {code}...')
				thisPromoCards = []
				for jsonCard in thisPromoJsonCards:
					thisPromoCards += Card.fromRawScryfall(self, jsonCard)
				for card in thisPromoCards:
					card.promoSetCode = code
				cards += thisPromoCards
			Indent.decrease()
		
		Indent.decrease()
		self.cards = cards
		if self.getImages:
			for card in self.cards:
				card.downloadImage()
	
	@property
	def tokens(self):
		return [card for card in self.cards if card.isToken]
	
	@property
	def lackeyCode(self):
		"""Uses the custom Lackey code if there is one, or else the standard"""
		if self.customCode is not None:
			return self.customCode
		return self.code
	
	def fetchToken(self, token, noHasteOkay=False):
		"""
		Return the correct card object to match the supplied (temp) token.
		Returns None if nothing matches.
		"""
		noHasteOkay = False
		if not self.tokensFetched:
			self.getCards(onlyTokens=True)
		display(2, f"Looking in {self} for a match to {token}")
		display(3, f"Color: '{token.color}'")
		display(3, f"Type: '{token.type_}'")
		display(3, f"Text: '{token.text}'")
		display(3, f"Power: '{token.power}'")
		display(3, f"Toughness: '{token.toughness}'")
		for myToken in self.tokens:
			if token.name == myToken.name:
				if token.matchOnName:
					display(2, f"Name match is good enough: {myToken}")
					return myToken
				Indent.increase()
				display(3, f"\nPotential match: {myToken}")
				display(3, f"Color: '{myToken.color}'")
				display(3, f"Type: '{myToken.type_}'")
				display(3, f"Text: '{myToken.text}'")
				display(3, f"Power: '{myToken.power}'")
				display(3, f"Toughness: '{myToken.toughness}'")
				Indent.decrease()
				if token == myToken:
					Indent.increase()
					display(2, f"\nFound a match: {myToken}\n")
					Indent.decrease()
					return myToken
				elif noHasteOkay and token.text == 'Haste':
					#Grumble grumble grumble. Wizards sometimes prints haste-only
					#tokens as vanilla and it's irritating.
					tokenCopy = copy(token)
					tokenCopy.text = ''
					if tokenCopy == myToken:
						Indent.increase()
						display(2, f"\nFound a vanilla match for haste-only: {myToken}")
						Indent.decrease()
						return myToken
		display(2, '\nNo match found.')
		return None
	
	def resolveSpawnScripts(self, review=False):
		"""Corrects set code for tokens not printed with this set."""
		display(3, f"Resolving spawn scripts for {self.name}")
		for card in self.cards:
			card.resolveSpawnScripts(review=review) #does nothing if card spawns no tokens
	
	def output(self):
		"""How to write this set to file"""
		if self.code == 'lackey':
			self.orderCards()
		
		output = ''
		if hasattr(self, 'cards'):
			for card in self.cards:
				output += card.output()
		else:
			raise AttributeError(f"Tried to output set {self.name}, but it doesn't have any cards.")
		output += '\n\n'
		return output
	
	def orderCards(self):
		""""Lets the custom Lackey token set order its tokens properly."""
		def nameSort(card):
			#if card.isCopy:
			#	return 'zzz' + card.lackeyName #Ensures copies are all last in each color
			return card.lackeyName
			
		cards = self.cards.copy()
		c = sorted([card for card in cards if card.color == '' and not card.isType('artifact') and not card.isType('counter')], key=nameSort)
		w = sorted([card for card in cards if card.color == 'W'], key=nameSort)
		u = sorted([card for card in cards if card.color == 'U'], key=nameSort)
		b = sorted([card for card in cards if card.color == 'B'], key=nameSort)
		r = sorted([card for card in cards if card.color == 'R'], key=nameSort)
		g = sorted([card for card in cards if card.color == 'G'], key=nameSort)
		m = sorted([card for card in cards if len(card.color) > 1], key=nameSort)
		a = sorted([card for card in cards if card.isType('artifact')], key=nameSort)
		counter = sorted([card for card in cards if card.isType('counter')], key=nameSort)
		cards = c + w + u + b + r + g + m + a + counter
		if len(cards) != len(self.cards):
			raise IndexError('Lackey set originally had {len(self.cards} tokens, but sorting somehow resulted in {len(cards)}.')
		self.cards = cards
	
	def __repr__(self):
		return f'<Set: {self.name}>'

class Card():
	"""A Magic card, or one face of a double-faced card."""
	
	cachedScripts = {}
	
	def __init__(self, name, set_, color, type_, cmc, rarity, **kwargs):
		"""
		Basic constructor. Take and assigns values.
		I'll need to call this sometimes outside of
		interpreting Scryfall JSON, so that process
		is a separate method that calls this.
		
		I'll never have to do math on any of these values,
		so it makes sense to store everything as strings.
		Defaults are '' instead of None. It's less idiomatic
		for Python, but it makes the JSON-handling and file-
		outputting code cleaner, and still evaulates to False.
		"""
		Indent.increase() #Manually, since the decorator breaks it
		self.name = name
		self.set_ = set_
		#Used for things like collector's boosters, where a promo card is included in a set, with a different Lackey code.
		self.promoSetCode = None
		self.color = color
		self.type_ = type_
		self.cmc = cmc
		self.number = number = kwargs.get('number', '')
		self.rarity = rarity
		self.sf_setCode = kwargs.get('sf_setCode', '')
		self.text = kwargs.get('text', '')
		self.power = kwargs.get('power', '')
		self.toughness = kwargs.get('toughness', '')
		self.cost = kwargs.get('cost', '')
		self.colorId = kwargs.get('colorId', '')
		#Any that are supplied:
		self._scripts = {
			'init': {
				'other': [],
				'token': [],
			},
			'callable': {
				'other': [],
				'token': [],
			},
		}
		self.draftQualities = kwargs.get('draftQualities', '')
		self.sound = kwargs.get('sound', '')
		self.imageFile = kwargs.get('imageFile', '')
		self.scryfall_image_uri = kwargs.get('scryfall_image_uri', '')
		self.loyalty = kwargs.get('loyalty', '')
		self.isToken = kwargs.get('isToken', False)
		self.isCopy = kwargs.get('isCopy', False)
		self.promoSetCode = kwargs.get('promoSetCode', None)
		
		Indent.decrease()
	
	def __repr__(self):
		return f"<Card '{self.name}' ({self.set_.code})>"
	
	def __eq__(self, other):
		"""Determine if two cards are functionally identical."""
		if all([
			self.name == other.name,
			self.color == other.color,
			self.type_ == other.type_,
			self.text == other.text,
			self.power == other.power,
			self.toughness == other.toughness,
		]):
			return True
		else:
			return False
	
	@property
	def scripts(self):
		return [script for timing in self._scripts.values() for effect in timing.values() for script in effect]
	
	@property
	def initScripts(self):
		return [script for effect in self._scripts['init'].values() for script in effect]
	
	@property
	def callableScripts(self):
		return [script for effect in self._scripts['callable'].values() for script in effect]
	
	@property
	def spawnScripts(self):
		return [script for timing in self._scripts.values() for script in timing['token'] ]
	
	def addScript(self, script):
		"""Adds a script, checking for duplicates and giving it a pointer pack to the card"""
		if isinstance(script, Iterable):
			for subitem in script:
				self.addScript(subitem)
			return
		timing = 'init' if script.init else 'callable'
		effect = 'token' if script.token else 'other'
		if script in self._scripts[timing][effect]:
			#Don't add a duplicate.
			return
		else:
			script.parentCard = self
			self._scripts[timing][effect].append(script)
	
	def generateScripts(self):
		"""
		Keep any scripts already supplied by the initial JSON parsing, and parse card text to generate more.
		Probably needs to be split up more at this point, as it's ballooned in size.
		"""
		if self.isToken is not True and Card.cachedScripts.get(self.name, None) is not None:
			#display(1, f'Using cached script for {self.name}')
			self.addScript(script.copy() for script in Card.cachedScripts[self.name])
			return
		
		if self.generateCustomScripts():
			return
		
		if self.text == '':
			return
		
		lowerText = self.text.lower()
		lowerName = re.escape(self.name.lower())
		
		#Flying. Find the word flying, in a line without a period
		if (self.isType('creature') or self.isType('vehicle')) and re.search(r'flying[^.]*?$', lowerText, flags=re.MULTILINE):
			self.addScript(Script(text='/ch'))
		
		#ETB tapped:
		if f'{self.name} enters the battlefield tapped' in self.text or 'it enters the battlefield tapped' in self.text or ('Myriad' in self.text and self.isToken):
			self.addScript(Script(text='/cr90'))
		
		if re.match(r'^Hideaway$', self.text, re.MULTILINE):
			self.addScript(Script(text='/cr90'))
			self.addScript(Script(label=f'Look at top 4', text=f'/vp1pt4'))
		
		#Doesn't untap:
		if re.search(r"doesn't untap during your untap step(?! if)", self.text):
			self.addScript(Script(text='/ci'))
		
		#Scrying, revealing, and looking at.
		scryMatches = re.findall(r'scry (\d+)', lowerText)
		if scryMatches:
			for number in scryMatches:
				self.addScript(Script(label=f'Scry {number}', text=f'/vp1pt{number}'))
		
		#Die-rolling
		rollMatch = re.search('rolls? (a|two|X) six-sided dic?e( twice)?', self.text, re.IGNORECASE)
		if rollMatch:
			display(2, 'Rolls dice')
			dieCount, twoDice = rollMatch.groups()
			display(2, (dieCount, twoDice))
			if dieCount == 'a' and not twoDice:
				display(2, 'Rolling one die')
				self.addScript(Script(label='Roll six-sided die', text='/roll 1'))
			elif dieCount == 'X':
				display(2, 'Rolling an indeterminate number of dice')
				self.addScript(Script(label='Roll six-sided die', text='/roll 1'))
				self.addScript(Script(label='Roll five six-sided die', text='/roll 5'))
				self.addScript(Script(label='Roll ten six-sided die', text='/roll 10'))
			else:
				display(2, 'Rolling 2 dice')
				self.addScript(Script(label='Roll two six-sided dice', text='/roll 2'))
		
		numbers = {
			'a': '1',
			'card': '1',
			'two': '2',
			'three': '3',
			'four': '4',
			'five': '5',
			'six': '6',
			'seven': '7',
			'eight': '8',
			'nine': '9',
			'ten': '10',
			'eleven': '11',
			'twelve': '12',
			'thirteen': '13'
		}
		reNumbers = '|'.join(numbers) #gets just the indices
		
		actionCodes = {
			'Look at': 'p',
			'Reveal': 'a',
		}
		lookRevealMatches = re.findall(rf'((?:looks? at)|(?:reveals?)|(?:exiles?)) the top ({reNumbers})', lowerText)
		if lookRevealMatches:
			for match in lookRevealMatches:
				action, numberWord = match
				action = action.capitalize()
				if action == 'Looks at':
					action = 'Look at'
				if action == 'Reveals':
					action = 'Reveal'
				if action == 'Exile' or action == 'Exiles':
					if 'face down' in self.text:
						action = 'Look at'
					else:
						action = 'Reveal'
				number = numbers[numberWord]
				if number == '1':
					strNumber = 'card'
				else:
					strNumber = numberWord + ' cards'
				code = actionCodes[action]
				self.addScript(Script(label=f'{action} top {strNumber}', text=f'/vp1{code}t{number}'))
		
		#Coin flipping
		flipMatch = re.search(f'flips? ({reNumbers}|X) coin', self.text, re.IGNORECASE)
		if flipMatch:
			numberWord = flipMatch.groups()[0]
			if numberWord == 'X':
				self.addScript(Script(label='Flip a coin', text='flip'))
				self.addScript(Script(label='Flip five coins', text='flip5'))
				self.addScript(Script(label='Flip ten coins', text='flip10'))
			else:
				number = numbers[numberWord]
				if number == '1':
					strNumber = 'a coin'
				else:
					strNumber = numberWord + ' coins'
				self.addScript(Script(label=f'Flip {strNumber}', text=f'/flip{number}'))
			
		
		#ETB with counters
		counterColor = None
		
		keywordCountersMatch = re.search(r'(?P<keyword>modular|graft|vanishing|fading) (?P<number>\d+)', lowerText)
		if keywordCountersMatch:
			keyword = keywordCountersMatch.group('keyword').capitalize()
			if keyword in ['Modular', 'Graft']:
				counterColor = 'green'
			else:
				counterColor = 'blue'
			counterNumber = keywordCountersMatch.group('number')
		elif self.isType('saga'):
			counterColor = 'blue'
			counterNumber = '1'
		elif self.loyalty and re.match(r'^\d+$', self.loyalty):
			counterColor = 'blue'
			counterNumber = self.loyalty
		elif self.name == 'Baron Von Count':
			counterColor = 'blue'
			counterNumber = '5'
		else:
			countersMatch = re.search(rf'(?<!if [^.]+?, ){lowerName} enters the battlefield (?:tapped (?:and )?)?with (?P<counterNumber>{reNumbers}) (?P<counterType>\S+?) counters? on it(?! for each)(?! if)', lowerText)
			if countersMatch:
				counterNumber = numbers[countersMatch.group('counterNumber')]
				counterType = countersMatch.group('counterType')
				if counterType == '+1/+1':
					counterColor = 'green'
				elif counterType == '-1/-1':
					counterColor = 'red'
				else:
					counterColor = 'blue'
		
		if counterColor:
			self.addScript(Script(text=f'/cc{counterColor}={counterNumber}'))
		
		#Keywords that can add a (usually fixed) number of +1/+1 counters
		addFixedMatch = re.search(r'(?P<keyword>monstrosity|adapt|renown|bloodthirst|tribute|fabricate) (?P<number>\d+)', lowerText)
		if addFixedMatch:
			keyword = addFixedMatch.group('keyword').capitalize()
			number = addFixedMatch.group('number')
			if number == '1':
				strNumber = 'a'
				plural = ''
			else:
				strNumber = number
				plural = 's'
			self.addScript(Script(label=f'{keyword} - Add {strNumber} +1/+1 counter{plural}', text=f'/ccgreen+{number}'))
		
		
		#(Deep breath) tokens, tokens, tokens...
		if 'create'.lower() in lowerText:
			display(3, 'This card references token creation.')
			
			#Catches and adds all "normal" token clauses.
			self.generateNormalSpawnScripts()
		
		#Various weird, named, or different spawn scripts:
		if self.isType('planeswalker') and 'emblem' in lowerText:
			#Spawn emblem.
			if self.name == 'Arlinn, Embraced by the Moon':
				emblemName = 'Arlinn Kord Emblem'
			else:
				emblemName = self.name + ' Emblem'
			lines = re.search(r'emblem with "(.+?)"(?: and "(.+?)")?', self.text).groups()
			text = '\n'.join(filter(None, lines))
			emblem = TempToken(emblemName, types='Emblem', text=text, matchOnName=True)
			self.addScript(Script(token=emblem))
		
		#Keyword mechanic abilities, spawned as "tokens":
		if not ' Counter' in self.name:
			counter_types = [
				'deathtouch',
				'double strike',
				'first strike',
				'flying',
				'hexproof',
				'indestructible',
				'lifelink',
				'menace',
				'reach',
				'trample',
				'vigilance',
			]
			for counter_type in counter_types:
				if re.search(f'{counter_type} counter', lowerText):
					if self.isType('instant') or self.isType('sorcery') or re.match(rf'When {self.name} enters the battlefield', self.text):
						init=True
					else:
						init=False
					self.addScript(Script.spawnCustomCounter(counter_type, init=init))
		
		#Token-spawning keyword abilities:
		servoMatch = re.search(r'Fabricate (\d+)', self.text)
		if servoMatch:
			#Spawn servo(s) for fabricate.
			qty = servoMatch.group(1)
			self.addScript(Script(token=TempToken('Servo', type_='Artifact Creature — Servo', power='1', toughness='1'), qty=qty))
		
		if 'investigate'.lower() in lowerText:
			clue = TempToken('Clue', type_='Artifact — Clue', text="{2}, Sacrifice this artifact: Draw a card.")
			#There are only two cards that currently investigate more than once, and no others investigate on ETB,
			#so I'm going to hard-code them for now. Can update if and when the mechanic returns.
			if self.name == 'Confirm Suspicions':
				self.addScript(Script(token=clue, qty='3', init=True))
			elif self.name == 'Declaration in Stone':
				self.addScript(Script(token=clue))
				self.addScript(Script(token=clue, qty='5'))
				self.addScript(Script(token=clue, qty='10'))
			else:
				if (self.isType('instant') or self.isType('sorcery')):
					self.addScript(Script(token=clue, init=True))
				else:
					self.addScript(Script(token=clue))
		
		if re.search('^Myriad$', self.text, re.MULTILINE):
			copyToken = TempToken.copyCard(self)
			self.addScript(Script.oneOrMoreSpawnScripts(copyToken, qty=1))
			self.addScript(Script.oneOrMoreSpawnScripts(copyToken, qty=2))
			self.addScript(Script.oneOrMoreSpawnScripts(copyToken, qty=3))
			self.addScript(Script.oneOrMoreSpawnScripts(copyToken, qty=4))
		if self.name == 'Blade of Selves':
			copyToken = TempToken('Copy')
			self.addScript(Script.oneOrMoreSpawnScripts(copyToken, qty=1))
			self.addScript(Script.oneOrMoreSpawnScripts(copyToken, qty=2))
			self.addScript(Script.oneOrMoreSpawnScripts(copyToken, qty=3))
			self.addScript(Script.oneOrMoreSpawnScripts(copyToken, qty=4))
			
		
		zombieMatch = re.search(r'^(Embalm|Eternalize).+$', self.text, re.MULTILINE)
		if zombieMatch:
			label = zombieMatch.group(1)
			type_ = self.type_.replace('—', '— Zombie')
			text = self.text.replace('\n' + zombieMatch.group(), '')
			if label == 'Eternalize':
				power = toughness = 4
				color = 'B'
			elif label == 'Embalm':
				power = self.power
				toughness = self.toughness
				color = 'W'
			zombie = TempToken(self.name, type_=type_, color=color, power=power, toughness=toughness, text=text, matchOnName=True)
			self.addScript(Script(token=zombie, label=label))	
		
		#"Tokens" for tracking player counters and similar things:
		if re.search(r'(?:\bInfect\b)|(?:poison counter)', self.text) and self.name != 'Leeches':
			self.addScript(Script(token=TempToken('Poison Counter', type_='')))
		if 'experience counter' in self.text:
			self.addScript(Script(label='Create Experience Counter', token=TempToken('Experience', type_='')))
		if '{E}' in self.text:
			self.addScript(Script(token=TempToken('Energy Reserve', type_='')))
		if re.search(r'^Ascend$', self.text, re.MULTILINE):
			self.addScript(Script(token=TempToken("City's Blessing", matchOnName=True, type_='')))
		if '— Adventure' in self.text:
			self.addScript(Script(label='Create Adventure Marker', token=TempToken('On an Adventure', matchOnName=True, type_='')))
		if self.name == 'Acornelia, Fashionable Filcher':
			self.addScript(Script(TempToken('Acorn Stash', matchOnName=True, type_='')))
	
	def generateCustomScripts(self):
		"""
		Generate custom scripts (and return True) for cards that DON'T have the word "create" in their text. Those are handled separately.
		"""
		if self.name == 'Kaya, Ghost Haunter':
			token1 = TempToken(name='Kaya, Ghost Haunter -1 Emblem', text='At the beginning of your upkeep, this emblem deals 3 damage to the owner of target haunted creature.')
			self.addScript(Script(token1))
			
			token2 = TempToken(name='Kaya, Ghost Haunter -2 Emblem', text='At the beginning of your upkeep, gain control of target haunted creature for as long as it remains haunted.')
			self.addScript(Script(token2))
			return True
		
		if self.name == 'Crystalline Giant':
			self.addScript(Script(label='Roll ten-sided die', text='/rolld10'))
			for number, counter_type in enumerate([
				'Flying',
				'First Strike',
				'Deathtouch',
				'Hexproof',
				'Lifelink',
				'Menace',
				'Reach',
				'Trample',
				'Vigilance',
			], start=1):
				self.addScript(Script.spawnCustomCounter(counter_type, label=f'({number}) Create {counter_type} Counter'))
			return True
		
		if self.name == 'Grimdancer':
			for counter_type in [
				'Menace',
				'Deathtouch',
				'Lifelink',
			]:
				self.addScript(Script.spawnCustomCounter(counter_type))
			return True
		
		elif self.name == 'Kathril, Aspect Warper':
			for counter_type in [
				'Flying',
				'First Strike',
				'Double Strike',
				'Deathtouch',
				'Hexproof',
				'Indestructible',
				'Lifelink',
				'Menace',
				'Reach',
				'Trample',
				'Vigilance',
			]:
				self.addScript(Script.spawnCustomCounter(counter_type))
			return True
		
		elif self.name == 'Goblin Bowling Team':
			self.addScript(Script(label='Roll six-sided die', text='/roll'))
			return True
		
		return False
	
	def generateNormalSpawnScripts(self):
		"""
		Find, process, and add all "normal" spawn scripts in a card's text.
		Is called on any card that contains the word "create"
		"""
		matchOnName = False
		
		#Catch custom scripts for specific cards.
		#The method adds them itself, and it returns
		#true if any were found.
		if self.generateCustomSpawnScripts():
			return
		
		numbers = {
			'X': 'X',
			'that many': 'X',
			'a number of': 'X',
			'a': '',
			'an': '',
			'two': '2',
			'three': '3',
			'four': '4',
			'five': '5',
			'six': '6',
			'seven': '7',
			'eight': '8',
			'nine': '9',
			'ten': '10',
			'eleven': '11',
			'twelve': '12',
			'thirteen': '13'
		}
		numbersPattern = r'(?P<number>{})'.format('|'.join(numbers)) #Gets just the indices
		
		#Cards that make copies of themselves
		copyMatch = re.search(rf"creates? {numbersPattern} tokens? (?:that's a copy|that are copies) of {re.escape(self.name)}", self.text, re.IGNORECASE)
		if copyMatch:
			qty = numbers[copyMatch.group('number')]
			self.addScript(Script.oneOrMoreSpawnScripts(token=TempToken.copyCard(self), qty=qty))
			return
		
		#Cards that make copies of arbitrary things.
		copyMatch = re.search(rf"creates? {numbersPattern} tokens? (?:that's a copy|that are copies) of", self.text, re.IGNORECASE)
		if copyMatch:
			qty = numbers[copyMatch.group('number')]
			self.addScript(Script.oneOrMoreSpawnScripts(token=TempToken('Copy'), qty=qty))
		
		#Grab all the (apparent) token clauses. Exit if none register.
		pattern = [
			rf"(?:^•.+?)?",
			rf"(?:if .*?)?",
			rf"(?:when {re.escape(self.name)} enters the battlefield, )?",
			rf"(?:For each .*?)?creates? .+? tokens?.*?",
			rf"(?:",
				r"\. Its power is equal to.+?and its toughness is equal to.+?\.",
				rf"|",
				rf"\".+?\"(?: and \".+?\")?",
				rf"|",
				rf"\.\"?(?:(?: It has| They have| They each have| The token has| Those tokens have).+?\.\"?)?",
				rf"|",
				rf"(?:create)",
			rf")",
		]
		pattern = ''.join(pattern)
		tokenMatches = re.findall(pattern, self.text, re.IGNORECASE|re.MULTILINE)
		if not tokenMatches:
			return
		
		#Generate the portions that we'll use in the regexes.
		regexParts = {}
		colors = '|'.join(('white', 'blue', 'black', 'red', 'green', 'colorless', 'gold', 'pink'))
		regexParts['colors'] = rf'(?P<color>(?:{colors})(?:, (?:{colors}))*(?:,? and (?:{colors}))?)'
		regexParts['numbers'] = numbersPattern
		regexParts['pt'] = r'(?P<power>[X\d]*)/(?P<toughness>[X\d]*)'
		regexParts['types'] = r'(?P<types>(artifact|creature|enchantment| )+)'
		
		#If this is a nonpermanent spell with only one summoning script, it should be automatic,
		#unless the spawning sentence includes the word "if".
		allInit = False
		if (self.type_ == 'Instant' or self.type_ == 'Sorcery') and len(tokenMatches) == 1:
			if not 'if ' in tokenMatches[0].casefold() and not '•' in tokenMatches[0].casefold():
				allInit = True
		
		tokenName = ''
		for tokenMatch in tokenMatches:
			display(3, f'TokenMatch: "{tokenMatch}"')
		
			#Should this script run as soon as the card is put on the table?
			init = False
			if allInit or (f'when {self.name} enters the battlefield'.casefold() in tokenMatch.casefold()):
				#This is a script that should run as the card is put on the table.
				init = True
		
			#Just use "Clone" button.
			#if tokenMatch == "a token that's a copy of " + name:
				#This makes a copy of itself.
			#	spawnScripts[init].append(spawnScript('Copy', name, init=init))
			#	continue
			
			#Capture the token's rules text, if any:
			keywords = ''
			quotedText = ''
			
			#First, keywords.
			keywordRegex = (
				#Keywords in the same sentence as "create":
				r"tokens? with ",
				rf"(?:",
					r"([^\".]+?)", #The list of keywords
					r"(?: that | for | if |\"|\.)",
					r"|",
					r"(?:(?: It has | They have )(.+?) and \")",
				r")",
				
				r"|",
				#Keywords in a separate sentence, like Rite of the Raging Storm or Estrid:				
				r"(?:\. It has |\. They have |\. They each have |\. The token has |\. Those tokens have )",
				r"([^\".]+?)", #The list of keywords
				r"(?: and \"|\.)", #End when we get to either quoted text or the end of the sentence. 
			)
			
			keywordRegex = ''.join(keywordRegex)
			keywordMatch = re.search(keywordRegex, tokenMatch, re.IGNORECASE)
			if keywordMatch:
				keywords = ' '.join(filter(None, keywordMatch.groups()))
				remove = re.search(r",? (?:where |that's |for |at |then |named |equal to|and roll ).*", keywords, re.IGNORECASE)
				if remove:
					keywords = keywords.replace(remove.group(), '')
				if ',' in keywords:
					sep = ','
				else:
					sep = ' and '
				keywords = [keyword.strip(', ') for keyword in keywords.split(sep)]
				keywords = [keyword.replace('and ', '') for keyword in keywords if keyword != 'and']
				keywords = ', '.join(filter(None, keywords)).lower().capitalize()
				display(3, f"keywords: '{keywords}'")
			
			#Next, quoted text.
			quotedTextRegex = (
				r"(?:with |and |\. It has |\. They have |\. The token has |\. Those tokens have )",
				r"(?:\"(.+?)\")",
				r"(?: and \"(.+?)\")?"
			)
			quotedTextRegex = ''.join(quotedTextRegex)
			quotedTextMatch = re.search(quotedTextRegex, tokenMatch, re.IGNORECASE)
			if quotedTextMatch:
				quotedText = '\n'.join(filter(None, quotedTextMatch.groups()))
				display(3, f"Quoted text: '{quotedText}'")
			
			text = '\n'.join(filter(None, (keywords, quotedText)))
			display(3, f"Text: '{text}'")
			
			
			pattern = (
				r"(?P<variable1>For each .*?)?",
				r"creates? {numbers}(?: tapped)?(?: legendary)?".format(**regexParts),
				rf"(?:",
					r"(?: (?P<predefined>treasure|food|gold) tokens?)",
					r"|",
					r"(?: {pt})?(?: {colors})?(?: (?P<subtypes>.+?))? {types} tokens?.*?".format(**regexParts),
				r")",
				r"(?P<wubrg> that's all colors)?(?P<variable2> for each)?",
				r"(?P<variablept>Its power is equal to.+?and its toughness is equal to.+?\.)?"
			)
			pattern = ''.join(pattern)
			match = re.search(pattern, tokenMatch, re.IGNORECASE)
			if match:
				#This appears to be a "normal" token clause.
				predefined = match.group('predefined')
				if predefined:
					#This is a token from the new "predefined" list.
					tokenName = predefined
					matchOnName = True
					types = ''
					subtypes = ''
					color = ''
					power = ''
					toughness = ''
					text = ''
				else:
					subtypes = match.group('subtypes')
					tokenName = subtypes
					if 'named' in tokenMatch:
						#Catches named tokens
						nameMatch = re.search(r"creates? .+?(?<!card )(?<!cards )named (?P<name>.+?)(?:\.| with| that's| at| for| and| from|, where| attached|$)", tokenMatch, re.IGNORECASE)
						if nameMatch:
							tokenName = nameMatch.group(1)
			
					power = match.group('power') or ''
					toughness = match.group('toughness') or ''
					types = match.group('types') or ''
					if 'creature' in types and power == '':
						power = toughness = '*'
					if match.group('wubrg'):
						color = 'WUBRG'
					else:
						color = extractColors(match.group('color'))
				
				display(4, f"Types: {types}")
				display(4, f"Subtypes: {subtypes}")
				display(4, f"Name: {tokenName}")
				token = TempToken(name=tokenName, types=types, subtypes=subtypes, color=color, power=power, toughness=toughness, text=text, matchOnName=matchOnName)
				display(3, f"Token: {token}")
				number = numbers[match.group('number')]
				if match.group('variable1') or match.group('variable2'):
					number = 'X'
				self.addScript(Script.oneOrMoreSpawnScripts(token, init=init, qty=number))
				continue
			
			match = re.search(r'{numbers} of those token'.format(**regexParts), tokenMatch)
			if match and token:
				#This is just a different number of whatever the previous token clause was.
				#Spawn that token, unless it wasn't specified.
				#(This should never be an initialization script.)
				number = numbers[match.group('number')]
				self.addScript(Script.oneOrMoreSpawnScripts(token, qty=number))
				continue
			
			match = re.search(r'create (?P<name>.+), a legendary {pt} {colors} (?P<subtypes>.+?) {types} '.format(**regexParts), tokenMatch, re.IGNORECASE)
			if match:
				#This is the new legendary template Verix Bladewing uses.
				tokenName = match.group('name')
				power = match.group('power')
				toughness = match.group('toughness')
				color = extractColors(match.group('color'))
				types = 'legendary ' + match.group('types')
				subtypes = match.group('subtypes')
				token = TempToken(name=tokenName, color=color, types=types, subtypes=subtypes, power=power, toughness=toughness, text=text)
				display(3, f"Token: {token}")
				self.addScript(Script(token=token, init=init))
				continue
	
	def generateCustomSpawnScripts(self):
		"""
		Generate custom scripts (and return True) for cards that DO have the word "create" in their text.
		"""	
		if self.name == 'Gunk Slug':
			self.addScript(Script(label='Create 3 Gunks', text='/spawnx3 [lackey] Gunk', init=True))
			return
		
		if self.name == 'Time Sidewalk':
			self.addScript(Script(label='Create 4 Time Walks', text='/spawnx4 Time Walk'))
			return True
		
		if self.name == 'Bone Rattler':
			self.addScript(Script(label='Create 4 Reassembling Skeletons', text='/spawnx4 Reassembling Skeleton'))
			return True
		
		if self.name == 'Generated Horizons':
			self.addScript(Script(label='Create Forest', text='/spawn Forest'))
			return True
		
		if self.name == 'Waste Land':
			self.addScript(Script(label='Create Wastes', text='/spawn Wastes'))
			return True
		
		if self.name == 'Pursued Whale':
			self.addScript(Script(TempToken('Pirate', color='R', power='1', toughness='1', text="This creature can't block.\nCreatures you control attack each combat if able.")))
			return True
		
		if self.name == 'Forbidden Friendship':
			self.addScript(Script(TempToken('Dinosaur', color='R', power='1', toughness='1', text='Haste'), init=True))
			self.addScript(Script(TempToken('Human Soldier', color='W', power='1', toughness='1'), init=True))
			return True
		
		if self.name == 'Bestial Menace':
			bestiary = (
				{'name': 'Snake', 'power': 1, 'toughness': 1},
				{'name': 'Wolf', 'power': 2, 'toughness': 2},
				{'name': 'Elephant', 'power': 3, 'toughness': 3},
			)
			for member in bestiary:
				self.addScript(Script(token=TempToken(**member, color='G'), init=True))
			return True
		
		
		if self.name == 'Broken Visage':
			self.addScript(Script(token=TempToken(name='Spirit', color='B', power='*', toughness='*')))
			return True
		
		if self.name == 'Chronozoa':
			#It's the only card that refers to itself as "it" while copying itself.
			self.addScript(Script(token=TempToken.copyCard(self), qty=2))
			return True
					
		if self.name == 'Jace, Cunning Castaway':
			illusion = TempToken('Illusion', color='U', power=2, toughness=2, text='When this creature becomes the target of a spell, sacrifice it.')
			self.addScript(Script(illusion))
			
			copy = TempToken.copyCard(self)
			copy.type_ = copy.type_.replace('Legendary ', '')
			self.addScript(Script(token=copy, qty=2))
			return True
		
		if self.name == 'Soul Separator':
			self.addScript(Script(token=TempToken('Copy')))
			self.addScript(Script(token=TempToken('Zombie', color='B', power='X', toughness='X')))
			return True
		
		if self.name == 'Rite of Replication':
			self.addScript(Script(token=TempToken('Copy')))
			self.addScript(Script(token=TempToken('Copy'), qty=5))
			return True
		
		if self.name == 'Rhys the Redeemed':
			#Technically makes copies, but only of tokens, so we don't want to spawn the copy token.
			self.addScript(Script(token=TempToken('Elf Warrior', color='GW', power=1, toughness=1)))
			return True
		
		if self.name == "Outlaws' Merriment":
			merryMen = (
				{'name': 'Human Warrior', 'power': 3, 'toughness': 1, 'text': 'Trample, haste'},
				{'name': 'Human Cleric', 'power': 2, 'toughness': 1, 'text': 'Lifelink, haste'},
				{'name': 'Human Rogue', 'power': 1, 'toughness': 2, 'text': 'Haste\nWhen this creature enters the battlefield, it deals 1 damage to any target.'},
			)
			for merryMan in merryMen:
				self.addScript(Script(token=TempToken(**merryMan, color='RW')))
			return True
		
		if self.name == 'Nahiri, the Lithomancer':
			self.addScript(Script(token=TempToken(name='Kor Soldier', color='W', power=1, toughness=1)))
			self.addScript(Script(token=TempToken(name='Stoneforged Blade', type_='Artifact — Equipment', text='Indestructible\nEquipped creature gets +5/+5 and has double strike.\nEquip {0}')))
			return True
		
		if self.name == 'Terastodon':
			#Technically variable, but it can only ever make up to three.
			token = TempToken(name='Elephant', color='G', power=3, toughness=3)
			self.addScript(Script(token=token))
			self.addScript(Script(token=token, qty=2))
			self.addScript(Script(token=token, qty=3))
			return True
		
		if self.name == 'Reef Worm':
			self.addScript(Script(token=TempToken(name='Fish', color='U', power=3, toughness=3, text='When this creature dies, create a 6/6 blue Whale creature token with "When this creature dies, create a 9/9 blue Kraken creature token."')))
			return True
			
		if self.name == 'Wurmcoil Engine':
			self.addScript(Script(label="Create Deathtouch Wurm", token=TempToken(name='Wurm', types='Artifact', power=3, toughness=3, text="Deathtouch")))
			self.addScript(Script(label="Create Lifelink Wurm", token=TempToken(name='Wurm', types='Artifact', power=3, toughness=3, text="Lifelink")))
			return True
		
		if self.name == 'Godsire':
			#This is the only damn card worded like this, so I'm hardcoding it.
			self.addScript(Script(token=TempToken(name='Beast', color='WRG', power=8, toughness=8)))
			return True
		
		if self.name == "Trostani's Summoner":
			#I'm not sure if anything else is worded this way, either.
			conclave = (
				{'name': 'Knight', 'color': 'W', 'power': 2, 'toughness': 2, 'text': 'Vigilance'},
				{'name': 'Centaur', 'color': 'G', 'power': 3, 'toughness': 3},
				{'name': 'Rhino', 'color': 'G', 'power': 4, 'toughness': 4, 'text': 'Trample'},
			)
			for member in conclave:
				self.addScript(Script(token=TempToken(**member), init=True))
			return True
		
		if self.name == 'Hazezon Tamar':
			#I'm spelling this out here and catching it in the table, JUST IN CASE
			#they ever actually print these.
			token = TempToken(name='Sand Warrior', color='RGW', power=1, toughness=1)
			self.addScript(Script(token=token))
			self.addScript(Script(token=token, qty=5))
			self.addScript(Script(token=token, qty=10))
			return True
		
		if self.name == 'Riptide Replicator':
			self.addScript(Script(label='Create token', token=TempToken(name='???', subtypes='???', color='WUBRG', power='*', toughness='*')))
			return True
		
		if self.name == "Volrath's Laboratory":
			self.addScript(Script(label='Create token', token=TempToken(name='???', subtypes='???', color='WUBRG', power='2', toughness='2')))
			return True
		
		if self.name == 'Sarpadian Empires, Vol. VII':
			bestiary = (
				{'name': 'Citizen', 'color': 'W', 'power': 1, 'toughness': 1},
				{'name': 'Camarid', 'color': 'U', 'power': 1, 'toughness': 1},
				{'name': 'Thrull', 'color': 'B', 'power': 1, 'toughness': 1},
				{'name': 'Goblin', 'color': 'R', 'power': 1, 'toughness': 1},
				{'name': 'Saproling', 'color': 'G', 'power': 1, 'toughness': 1},
			)
			for entry in bestiary:
				self.addScript(Script(token=TempToken(**entry)))
			return True
		
		if self.name == 'Master of the Wild Hunt Avatar':
			bestiary = (
				{'name': 'Wolf', 'color': 'G', 'power': 2, 'toughness': 2},
				{'name': 'Antelope', 'color': 'G', 'power': 2, 'toughness': 3, 'text': 'Forestwalk'},
				{'name': 'Cat', 'color': 'G', 'power': 3, 'toughness': 2, 'text': 'Shroud'},
				{'name': 'Rhino', 'color': 'G', 'power': 4, 'toughness': 4, 'text': 'Trample'},
			)
			for entry in bestiary:
				self.addScript(Script(token=TempToken(**entry)))
			return True
		
		if self.name == 'Homarid Spawning Bed':
			#The default script finds this just fine, buuuut I want to be able to make exactly
			#8, for Scornful Egotist and Maelstrom Djinn. : )
			self.addScript(Script(token=TempToken(name='Camarid', color='U', power=1, toughness=1)))
			self.addScript(Script(token=TempToken(name='Camarid', color='U', power=1, toughness=1), qty=5))
			self.addScript(Script(token=TempToken(name='Camarid', color='U', power=1, toughness=1), qty=8))
			self.addScript(Script(token=TempToken(name='Camarid', color='U', power=1, toughness=1), qty=10))
			return True
		
		if self.name == 'Sword of Dungeons & Dragons':
			#It would catch the token just fine, but this adds the die roll too.
			self.addScript(Script(label='Roll d20', text='/rolld20'))
			self.addScript(Script(TempToken('Dragon', color='Gold', power=4, toughness=4, text='Flying')))
			return True
		
		return False
	
	def resolveSpawnScripts(self, review=False):
		"""Update set code for tokens not printed with this card's set."""
		pass
		for script in self.spawnScripts:
			script.resolve(review=review)
	
	def outputScripts(self):
		#Right now, scripts don't support spawning multiple tokens each, so these have to be hacked in manually on output.
		if self.name == 'Wurmcoil Engine':
			tokens = [script.token for script in self.spawnScripts]
			spawns = ';'.join([f'/spawn {token.lackeyName}' for token in tokens])
			return f'<s><l>Create Wurms</l><f>{spawns}</f></s>'
		
		if self.name == 'Soul Separator':
			tokens = [script.token for script in self.spawnScripts]
			spawns = ';'.join([f'/spawn {token.lackeyName}' for token in tokens])
			return f'<s><l>Create Copy and Zombie</l><f>{spawns}</f></s>'
		
		initScriptString = ';'.join([script.output() for script in self.initScripts])
		
		if len(self.callableScripts) == 0:
			return initScriptString
		else:
			output = ''
			if initScriptString:
				output = f'<s><a>y</a><l>Initialize</l><f>{initScriptString}</f></s> '
			
			if self.name == 'Kathril, Aspect Warper':
				tokens = [script.token for script in self.spawnScripts]
				spawns = ';'.join([f'/spawn {token.lackeyName}' for token in tokens])
				output += f'<s><l>Create all 11 counters</l><f>{spawns}</f></s> '
			
			output += ' '.join([script.output() for script in self.callableScripts])
			
			return output
	
	def output(self):
		"""Output to be written to Lackey data file."""
		#Exception for promos included in the set, like Collector's Booster cards, that have a separate Lackey code
		if self.promoSetCode:
			setCode = self.promoSetCode
		else:
			setCode = self.set_.lackeyCode
		
		fields = [
			self.lackeyName,
			setCode,
			self.imageFile,
			self.color,
			self.colorId,
			self.cost,
			self.cmc,
			self.type_,
			self.power,
			self.toughness,
			self.loyalty,
			self.rarity,
			self.draftQualities,
			self.sound,
			self.outputScripts(),
			self.text,
		]
		fields = [escape(field) for field in fields]
		return ('\t'.join(fields) + '\n')
	
	@property
	def lackeyName(self):
		"""Return the name to be recorded for Lackey. Only different for tokens."""
		
		if self.isToken:
			if self.set_ == None:
				return ''
			if self.isToken == 'quasi':
				#Backside or flip, not an actual token
				return f'[{self.set_.code}] {self.name}'
			elif self.isCopy:
				return f'[{self.set_.code}] {self.name} (Copy)'
			else:
				outputName = self.name
				parenthetical = []
				if self.color != '':
					parenthetical.append(self.color)
				power = self.power
				toughness = self.toughness
				if power != '':
					if power == 'X':
						power = '*'
					if toughness == 'X':
						toughness = '*'
					parenthetical.append(f'{power}/{toughness}')
					
					commaItems = []
					if 'Enchantment' in self.type_ and 'Creature' in self.type_:
						commaItems.append('Enchantment')
					
					keywords = re.findall(r'^([^.]+?)\n?$', self.text, re.MULTILINE) #If there are any lines with no periods, it's just keywords
					if keywords:
						keywords = tuple('Bands with Other' if 'bands with other'.casefold() in keyword.casefold() else keyword for keyword in keywords)
						commaItems += keywords
						#parenthetical.append(', '.join(keywordsMatches).title())
					
					if commaItems:
						parenthetical.append(titlecase(', '.join(commaItems)))
				if parenthetical:
					parenthetical = ' '.join(parenthetical)
					outputName = f'{self.name} ({parenthetical})'
				
				return f'[{self.set_.code}] {outputName}'
		
		return self.name
	
	@indented
	def downloadImage(self):
		"""Download this card's image, prepending temp or partial as needed."""
		display(3, f'{self.name} URI: {self.scryfall_image_uri}')
		Indent.increase()
		response = requests.get(self.scryfall_image_uri)
		img = Image.open(BytesIO(response.content))
		
		display(3, 'Resizing...')
		target_width, target_height = 312, 445
		target_aspect =	 target_width / target_height
		source_aspect = img.width / img.height
		
		if source_aspect == target_aspect:
			img = img.resize((target_height, target_height), Image.LANCZOS)
		
		elif source_aspect > target_aspect:
			#Source image is too wide. Resize based on height, then crop the sides.
			new_width = round(target_height / img.height * img.width)
			img = img.resize((new_width, target_height), Image.LANCZOS)

			# border_left = round((img.width - target_width) / 2)
			# error = border_left * 2 + target_width - img.width
			# border_right = border_left - error
			
			error = img.width - target_width
			border_left = round(error / 2)
			border_right = error - border_left
			
			img = img.crop((border_left, 0, img.width - border_right, img.height))
		elif source_aspect < target_aspect:
			#Source image is too tall. Resize based on width, then crop the top and bottom.
			new_height = round(target_width / img.width * img.height)
			img = img.resize((target_width, new_height), Image.LANCZOS)
			
			# border_top = round((img.height - target_height) / 2)
			# error = border_top * 2 + target_height - img.height
			# border_bottom = border_top - error
			
			error = img.height - target_height
			border_top = round(error / 2)
			border_bottom = error - border_top
			
			img = img.crop((0, border_top, img.width, img.height - border_bottom))
			
		if img.size != (target_width, target_height):
			raise ValueError(f"Somehow {self.name}'s image, after resizing and cropping, was {img.width} x {img.height} pixels. (Goal: {target_width} x {target_height})")
		
		if 'flipped' in self.number:
			display(3, "Ah, a flip card's upside-down form.")
			img = img.rotate(180)
		
		filename = os.path.join(imagesDir, escape(self.imageFile) + '.jpg')
		
		os.makedirs(os.path.dirname(filename), exist_ok=True)
		display(3, f'Saving as {filename}')
		img.save(filename, quality=85)
		Indent.decrease()
	
	@indented
	def createTokenImage(self):
		"""
		Create an image for a token that doesn't have an official version in the plugin.
		"""
		tokenImage = TokenImage(self)
		tokenImage.save()
	
	def tokenImageFileName(self, includeFolder=True, includeVersion=True):
		filename = self.lackeyName
		replacements = [
			(r'[(),{}]', ''),
			(r'\[lackey\] ', ''),
			(r'\*', 'X'),
			(r'[\s/?]', '-')
		]
		for regex, replace in replacements:
			filename = re.sub(regex, replace, filename) #Strip off set code and slugify
		if includeVersion:
			if self.isType('enchantment') and self.isType('creature'):
				filename += '-v5'
			elif self.isType('counter'):
				filename += '-v2'
			else:
				filename += '-v4'
		if includeFolder:
			filename = 'lackey/' + filename
		return filename.lower()
	
	@classmethod
	@indented
	def fromRawScryfall(cls, set_, jsonCard):
		"""
		Take the set's self-reference and JSON data
		from Scryfall and return
		a LIST of card objects (one for a normal card,
		two for a double-faced card)
		"""
		returnCards = []
		layout = jsonCard['layout']
		number = jsonCard['collector_number']
		
		if layout == 'transform':
			#If it's a transform card, split it into JSON cards.
			cmc = jsonCard['cmc']
			front = jsonCard['card_faces'][0]
			back = jsonCard['card_faces'][1]
			front['color_identity'] = jsonCard['color_identity']
			front['booster'] = jsonCard['booster']
			front['set'] = back['set'] = jsonCard['set']
			front['cmc'] = back['cmc'] = jsonCard['cmc']
			front['rarity'] = back['rarity'] = jsonCard['rarity']
			front['highres_image'] = back['highres_image'] = jsonCard['highres_image']
			front['collector_number'] = number + 'a'
			back['collector_number'] = number + 'b'
			front['scripts'] = [ Script(token=TempToken(back['name'], isToken='quasi'), label='Transform') ]
			back['layout'] = 'backside' #Custom layout name, not from Scryfall
			return [
				cls.fromSplitScryfall(set_, front),
				cls.fromSplitScryfall(set_, back)
			]
		elif layout == 'meld':
			#Meld is freaking weird.
			if jsonCard['mana_cost'] == '' and not 'Land' in jsonCard['type_line']:
				#This is the backside. Treat like the backside of a transform card.
				jsonCard['layout'] = 'backside'
				return [cls.fromSplitScryfall(set_, jsonCard)]
			else:
				#This is the front side. Give it a transform script and send it on its way.
				meldName = False
				for item in jsonCard['all_parts']:
					if item['component'] == 'meld_result':
						meldName = item['name']
						break
				if not meldName:
					raise ValueError(f"{jsonCard['name']} appears to be a meld card, but its back side can't be found.")
				jsonCard['scripts'] = [ Script(token=TempToken(meldName, isToken='quasi'), label=f'Meld into {meldName}') ]
				return [cls.fromSplitScryfall(set_, jsonCard)]
		elif layout == 'double_faced_token':
			#If it's a two-sided token, just use the front.
			cmc = jsonCard['cmc']
			front = jsonCard['card_faces'][0]
			front['set'] = jsonCard['set']
			front['cmc'] = jsonCard['cmc']
			front['highres_image'] = jsonCard['highres_image']
			front['rarity'] = jsonCard['rarity']
			front['layout'] = 'token'
			front['collector_number'] = number
			return [cls.fromSplitScryfall(set_, front)]
		elif layout == 'flip':
			#Generate a script and extra "side" so that
			#flip cards can operate like DFCs.
			top = jsonCard['card_faces'][0]
			bottom = jsonCard['card_faces'][1]
			bottom['collector_number'] = number + 'flipped'
			bottom['layout'] = 'bottom' #Custom, non-Scryfall layout value
			bottom['set'] = jsonCard['set']
			bottom['cmc'] = jsonCard['cmc']
			bottom['rarity'] = jsonCard['rarity']
			bottom['image_uris'] = jsonCard['image_uris']
			bottom['highres_image'] = jsonCard['highres_image']
			jsonCard['name'] = top['name']
			jsonCard['oracle_text'] = f"{top['oracle_text']}\n//\n[{bottom['name']}] {bottom['oracle_text']}"
			jsonCard['scripts'] = [ Script(token=TempToken(bottom['name'], isToken='quasi'), label='Flip') ]
			return [
				cls.fromSplitScryfall(set_, jsonCard),
				cls.fromSplitScryfall(set_, bottom)
			]
		elif layout == 'adventure':
			#So far, at least, Scryfall stores these like split cards. I don't want to.
			fields = ('name', 'type_line', 'mana_cost')
			creature, adventure = jsonCard['card_faces']
			
			for field in fields:
				jsonCard[field] = creature[field]
			jsonCard['oracle_text'] = f"{creature['oracle_text']}\n//\n{adventure['name']} | {adventure['mana_cost']} | {adventure['type_line']} | {adventure['oracle_text']}"
			return [cls.fromSplitScryfall(set_, jsonCard)]
		else:
			#Ah, a normal Magic card (or a split card). Just get its info.
			return [cls.fromSplitScryfall(set_, jsonCard)]
	
	@classmethod
	def fromSplitScryfall(cls, set_, jsonCard):
		"""
		An intermediary between fromRawScryfall and the
		plain constructor. Return a single card object.
		"""
		name = jsonCard['name']
		number = jsonCard['collector_number']
		display(3, f'Parsing #{number}: {name}...')
		Indent.increase()
		sf_setCode = jsonCard['set'] #This is the code as it's listed on Scryfall.
		#Define WUBRG order, mostly to get canonical multicolor token names
		color = ''.join(orderColors(jsonCard.get('colors', [])))
		colorId = ''.join(orderColors(jsonCard.get('color_identity', [])))
		#cost = jsonCard.get('mana_cost', '').replace('{', '').replace('}', '')
		#cost = simplifyCost(jsonCard.get('mana_cost', ''))
		cost = jsonCard.get('mana_cost', '')
		cmc = jsonCard['cmc']
		loyalty = jsonCard.get('loyalty', '')
		if cmc == 0.5:
			cmc = '1/2'
		else:
			cmc = str(int(cmc))
		type_ = jsonCard['type_line']
		power = jsonCard.get('power', '')
		toughness = jsonCard.get('toughness', '')
		if 'Basic Land' in type_ and name != 'Wastes':
			rarity = 'L'
		else:
			rarity = jsonCard['rarity'][0].upper()
		number = jsonCard['collector_number']
		
		layout = jsonCard.get('layout', '')
		if sf_setCode in ('tfth', 'tbth', 'tdag', 'thp3'): #The "tokens" in the challenge decks aren't really tokens. Neither are the enhancement stickers.
			layout = 'normal'
			type_ = type_.replace('Token ', '')
			if name == 'Enhancement Stickers':
				type_ = 'Hero Stickers'
		if layout == 'split':
			text = ' // '.join([face['oracle_text'] for face in jsonCard['card_faces']])
		else:
			text = jsonCard.get('oracle_text', '') #Don't replace newlines yet before parsing scripts.
		if text:
			text = re.sub(r'[ ]?\(.+?\)', '', text) #Remove reminder text, including the space before it
			#Removing leading newline, if the card's text began with reminder text. Avoid fucking up adventurers.
			if text and text[0] == '\n' and text[:3] != '\n//':
				text = text[1:]
		
		if type_ == 'Vanguard':
			text += f" | Hand Size {jsonCard['hand_modifier']} | Starting Life {jsonCard['life_modifier']}"
		
		scryfall_image_uri = jsonCard['image_uris']['border_crop']
		prepend = ''
		if set_.temp or number in set_.tempNumbers:
		#if not jsonCard['highres_image']: #Apparently this value is not consistently supported or trustable.
			prepend += 'temp_'
			if number in set_.partials:
				prepend += 'partial_'
		if sf_setCode == 'con':
			#Custom folder name for Conflux, because it's a protected file name under Windows.
			#(Tokens still go in tcon.)
			imageFile = os.path.join('cfx', prepend + number)
		else:
			imageFile = os.path.join(sf_setCode, prepend + number)
		
		#Custom hacks:
		if name == 'Kaya, Ghost Haunter':
			text = text.replace('emblem with,', 'emblem with')
		if name == 'B.F.M. (Big Furry Monster)' and number == '29':
			name = 'B.F.M. (Big Furry Monster) (Right Side)'
		elif sf_setCode == 'tisd' and name == 'Ooze':
			text = '' #Scryfall gives it the text it's SUPPOSED to have, but the printed token doesn't have it.
		elif (sf_setCode in ['tust', 'h17', 'tund']) and name == 'Dragon':
			color = 'Gold' #heeheehee
		elif name == 'Giant Teddy Bear':
			color = 'Pink'
		elif name == 'Feather':
			text = '{1}, Sacrifice Feather: Return target Phoenix card from your graveyard to the battlefield tapped.'
		elif name == 'Dinosaur Cat':
			type_ = 'Token Creature — Dinosaur Cat'
		elif name == 'Whirlwind of Thought':
			cmc = '4'
		elif name == 'Pirate' and sf_setCode == 'tm21':
			power = '1'
			toughness = '1'
		
		data = {
			'name': name,
			'set_': set_,
			'sf_setCode': sf_setCode,
			'color': color,
			'type_': type_,
			'cmc': cmc,
			'number': number,
			'rarity': rarity,
			'text': text,
			'power': power,
			'toughness': toughness,
			'cost': cost,
			'colorId': colorId,
			'imageFile': imageFile,
			'scryfall_image_uri': scryfall_image_uri,
			'loyalty': loyalty,
		}
		
		card = Card(**data)
		
		#Handling if this card is a token. Typeless cards printed as tokens,
		#like City's Blessing and Energy Reserve, have the 'normal' layout
		#and a type line that just reads 'card'. Backs of DFC are sort of tokens,
		#but they should keep their rarity and don't need color, power, and toughness
		#in their name.
		
		if layout == 'backside' or layout == 'bottom':
			card.isToken = 'quasi' #Treated like a token, but actually part of a card
		elif layout == 'token' or layout == 'emblem' or type_ == 'Card' or name == 'Morph':
			card.isToken = True
			card.colorId = '' #Tokens shouldn't have a color ID.
			card.rarity = 'T'
			#Emblems are no longer printed with subtypes.
			#I'm updating old ones.
			if layout == 'emblem':
				type_ = 'Emblem'
		
		#Add "token"  to type line if it doesn't already have it (for emblems, for instance).
		if card.isToken and not 'Token' in type_:
			card.type_ = 'Token ' + type_
			if card.type_ == 'Token Card':
				card.type_ = 'Token'
		
		#We have to make sure "token" is first, before "legendary".
		card.type_ = card.type_.replace('Legendary Token', 'Token Legendary')
		
		if set_.code == 'ust':
			#Add the variant number to get unique names.
			match = re.search(r'[bcdef]', jsonCard['collector_number'])
			if match:
				card.name += ' ({})'.format(match.group())
				display(3, f'Changing name to {card.name}')
		
		for script in jsonCard.get('scripts', []):
			card.addScript(script)
		card.generateScripts()
		if card.isToken is not True and Card.cachedScripts.get(card.name, None) is None:
			Card.cachedScripts[card.name] = card.scripts
		if card.scripts:
			for script in card.scripts:
				display(3, script.output())
		
		if not card.isToken:
			if set_.isDrafted is not False:
				#Build a field of qualities that only matter (to Lackey) for pack-building.
				#If the set doesn't say it's not drafted, but the category is by default not drafted,
				#the code doesn't know it yet, so it goes ahead and adds "notinPacks" to everything.
				#That's removed afterwards by the category's getCards method.
				draftQualities = []
				
				if not jsonCard['booster']:
					draftQualities.append('notInPacks')
				
				if set_.draftQualities:
					for quality in set_.draftQualities:
						if quality == 'watermark' and jsonCard.get('watermark'):
							draftQualities.append(jsonCard['watermark'])
						if quality == 'colorshifted' and 'colorshifted' in jsonCard.get('frame_effects', []):
							draftQualities.append('colorshifted')
						if quality == 'tapland':
							if card.rarity == 'C' and card.type_ == 'Land' and card.name != 'Evolving Wilds':
								draftQualities.append('tapland')
		
				#SeededPacks are defined in a different format.
				seededPackNames = SeededPackRecord.getPackNamesForCard(card)
				if seededPackNames:
					display(3, f"In seeded pack(s): {', '.join(seededPackNames)}")
					draftQualities += seededPackNames
		
				card.draftQualities = ' '.join(draftQualities)
			
			#Set sounds:
			# if 'Creature' in card.type_:
			# 	card.sound = 'creature'
			# elif 'Artifact' in card.type_:
			# 	card.sound = 'artifact'
			# elif 'Enchantment' in card.type_:
			# 	card.sound = 'enchantment'
			# elif 'Sorcery' in card.type_:
			# 	card.sound = 'sorcery'
			# elif 'Instant' in card.type_:
			# 	card.sound = 'instant'
			# elif 'Basic' in card.type_ and name != 'Wastes':
			# 	card.sound = card.name.lower()
			# elif 'Land' in card.type_:
			# 	card.sound = 'land'
			# elif 'Planeswalker' in card.type_:
			# 	card.sound = 'planeswalker'
			card.sound = ''
		
		Indent.decrease()
		
		return card
	
	@classmethod
	def fromToken(cls, token, set_, image=True):
		"""Create a card object for the custom Lackey tokens set."""
		card = Card(
			name=token.name,
			set_=set_,
			imageFile='', #Has to be set after instantiation
			color=token.color,
			type_=token.type_,
			power=token.power,
			toughness=token.toughness,
			loyalty=token.loyalty,
			text=token.text,
			cost=token.cost,
			cmc=token.cmc,
			rarity='T',
			isToken=token.isToken,
			isCopy=token.isCopy,
		)
		card.generateScripts()
		#Uncomment to save custom token images!
		#card.createTokenImage()
		if image:
			card.imageFile = card.tokenImageFileName()
		else:
			card.imageFile = ''
		
		return card
	
	def asTuple(self):
		"""God, Card and TempToken reeeaaally need to co-inherit from something already."""
		return (
			self.name,
			self.type_,
			#Cost and CMC are only used for self-copies, which shouldn't ever need this.
			self.color,
			self.power,
			self.toughness,
			re.sub(r'\s', '+', self.text).casefold()
		)
	
	def isType(self, type_):
		return type_.lower() in self.type_.lower()

class Script():
	"""A script on a card."""
	
	def __init__(self, token=None, init=None, qty='', text=None, label=None):
		self.token = token
		self.text = text
		self.label = label
		self.qty = str(qty)
		if self.qty == '1':
			self.qty = ''
		self.init = init
		
		if self.init is None:
			if self.text and not self.label:
				self.init = True
		
		if self.token and not self.init and not self.label:
			labelName = self.token.name
			if self.token.isCopy:
				labelName += ' Copy'
			if self.qty:
				custom = False
				plurals = {
					('Elf', 'Elves'),
					('Dwarf', 'Dwarves'),
					('Sand', 'Sand'),
					('Pegasus', 'Pegasi'),
					('Homunculus', 'Homunculi'),
					('Wolf', 'Wolves'),
					('Copy', 'Copies'),
					('Fish', 'Fish'),
					('Starfish', 'Starfish'),
					('Half', 'Halves'),
				}
				#Pluralize with custom table, 'es', or 's'.
				for singular, plural in plurals:
					if labelName[-len(singular):] == singular:
						labelName = labelName[:-len(singular)] + plural
						custom = True
						break
				if custom == False:
					if labelName[-1:] == 's':
						labelName = labelName + 'es'
					else:
						labelName = labelName + 's'
			self.label = ' '.join(filter(None, ('Create', self.qty, labelName)))
	
	@classmethod
	def oneOrMoreSpawnScripts(cls, token, qty, init=None, text=None, label=None):
		"""
		Returns a tuple containing one or more spawn scripts;
		One if qty is fixed, or version with qty=1, 5, and 10 if variable.
		"""
		if qty == '' or type(qty) == int or (type(qty) == str and qty.isdigit()):
			return (cls(token=token, qty=qty, init=init, text=text, label=label),)
		if qty == 'X':
			return (
				cls(token=token, qty=1, init=False, text=text, label=label),
				cls(token=token, qty=5, init=False, text=text, label=label),
				cls(token=token, qty=10, init=False, text=text, label=label),
				)
		raise ValueError(f'Quantity must be blank, a number, or X. It was: {qty}')
	
	@classmethod
	def spawnCustomCounter(cls, counter_type, label=None, init=False):
		counter_type = counter_type.title()
		text = {
			'Deathtouch': "(Any amount of damage dealt by a permanent with deathtouch to a creature is enough to destroy it.)",
			'First Strike': "(A creature with first strike deals damage before creatures without first strike.)",
			'Double Strike': "(A creature with double strike deals both first-strike and regular damage.)",
			'Flying': "(A creature with flying can't be blocked except by creatures with flying or reach.)",
			'Hexproof': "(A permanent with hexproof can't be the target of spells or abilities your opponents control.)",
			'Indestructible': "(Damage and effects that say \"destroy\" don't destroy a permanent with indesctructible.)",
			'Lifelink': "(Damage dealt by a permanent with lifelink also causes you to gain that much life.)",
			'Menace': "(A creature with menace can't be blocked except by two or more creatures.)",
			'Reach': "(A creature with reach can block creatures with flying.)",
			'Trample': "(A creature with trample can deal excess combat damage to the player or planeswalker it's attacking.)",
			'Vigilance': "(Attacking doesn't cause a creature with vigilance to attack.)",
		}
		#text = f"(A creature with one or more {counter_type.lower()} counters on it has {counter_type.lower()}.)"
		token = TempToken(f'{counter_type} Counter', type_='Token Counter', text=text[counter_type], matchOnName=True)
		return cls(token, label=label, init=init)
	
	def __eq__(self, other):
		if all([
			self.token == other.token,
			self.text == other.text,
			self.label == other.label,
			self.qty == other.qty,
			self.init == other.init,
		]):
			return True
		else:
			return False
	
	def __repr__(self):
		if self.label:
			labelDisplay = f"Label: '{self.label}', "
		else:
			labelDisplay = ''
		
		if self.token:
			return f"<Script: {labelDisplay}Spawn: '{self.token.name}'>"
		elif self.text:
			return f"<Script: {labelDisplay}Text: '{self.text}'>"
	
	@classmethod
	def resolveAllScripts(cls):
		for category in Category.instancesList():
			category.resolveSpawnScripts()
	
	def resolve(self, review=False):
		"""
		Decide from which set this scripts' token should be spawned.
		The token-class object is only a placeholder. This will
		replace it with a reference to the proper card-class object.
		"""
		display(3, f'\nResolving a script for #{self.parentCard.number}. {self.parentCard.name}')
		if self.token and type(self.token) == TempToken: #Must spawn a token, and must not be resolved yet
			
			#1) Check if the correct token exists in this set's tokens.
			if not self.token.isCopy: #Token copies are all custom-made (so far).
				thisSetToken = self.set_.fetchToken(self.token, noHasteOkay=True)
				if thisSetToken:
					self.token = thisSetToken
					return
				else:
					display(3, f'{self.token} not found in its own set {self.set_}')
			
				#2) Check elsewhere in the block.
				if hasattr(self.set_, 'block'):
					#Suppress the million messages we'll get from this.
					#savedVerbosity = verbosity_setting.verbosity
					#verbosity_setting.verbosity = 0
					for set_ in self.set_.block.sets:
						if set_ is not self.set_:
							foundToken = set_.fetchToken(self.token, noHasteOkay=True)
							if foundToken:
								self.token = foundToken
								display(3, f"{self} in {self.set_}: set changed to {set_}, also in the {self.set_.block} block")
								#Restore verbosity.
								#verbosity_setting.verbosity = savedVerbosity
								return
						display(3, f'Not found in its block either. {self.set_.block}')
					#Restore verbosity.
					#verbosity_setting.verbosity = savedVerbosity
			
			if review:
				#3, option 1) At this point we have a missed token. We need to create (or find) a missed record. Also creates a new custom token if need be.
				missedRecord = MissedTokenRecord.findOrCreate(self.token, createImage=self.set_.code!='cmb1')
				
				#Adds the current set, if it's not already in the list of sets that need this token. 
				missedRecord.register(self.set_)
			
				#Assign the custom token, or nab one of the available tokens (will later add functionality to save preferences).
				if missedRecord.customToken:
					display(3, f"No match found for {self.token} spawned by {self.parentCard}, so using custom token")
					self.token = missedRecord.customToken
				else:
					#If only one set has it, use that. Otherwise, leave the script unresolved for now.
					if len(missedRecord.setsThatHave) == 1:
						setThatHas = missedRecord.setsThatHave[0]
						display(3, f"There wasn't a rewrite rule for the {self.token} spawned by {self.parentCard}, but {setThatHas} is the only set that has it.")
						self.token = setThatHas.fetchToken(self.token)
				return
			
			#3, option 2) Check if there's already a saved rule for this.
			rerouteRule = TokenRerouteRule.findForToken(self.token)
			if rerouteRule:
				if rerouteRule.defaultSetCode == 'lackey':
					#Save this card for use in the token image gallery.
					if not self.parentCard.isCopy:
						LackeyOriginRecord.register(self.token, self.parentCard)
					#If we're using a custom token, but it doesn't exist yet, generate it.
					lackey = Set['lackey']
					if not lackey.fetchToken(self.token):
						if self.set_.code == 'cmb1':
							createImage = False
						else:
							createImage = True
						lackey.cards.append(Card.fromToken(self.token, lackey, image=createImage))
				rerouteRule.replaceTokeninScript(self) #Replace it with the correct card object.
				display(3, f"\t{self} in {self.set_}: set changed to {self.token.set_} due to preset rule")
				if not review:
					#Then if we're not reviewing spawn reroute rules, we should be done here.
					return
			
			if self.set_.spawnBlanks:
				#Partially spoiled set, without tokens yet. Spawn a blank card.
				display(3, 'Setting to blank spawn.')
				self.token = Card.fromToken(self.token, None)
				return
			
			#If there was no applicable reroute rule, raise an error.
			raise KeyError(f'No existing reroute rule found for {self.token}. Did you want to run in token-review mode?')
	
	def copy(self):
		if self.token is None:
			return self
		else:
			return Script(token=copy(self.token), init=self.init, qty=self.qty, label=self.label)
	
	@property
	def set_(self):
		return self.parentCard.set_
	
	def output(self):
		if self.text:
			command = self.text
		else:
			if self.qty:
				strQty = 'x' + self.qty
			else:
				strQty = ''
			#print(self.parentCard)
			command = ' '.join(filter(None, (f'/spawn{strQty}', self.token.lackeyName)))
		
		if self.init:
			return command
		else:
			return f'<s><l>{self.label}</l><f>{command}</f></s>'

class TempToken():
	"""A placeholder for spawn scripts to use, before they resolve to actual cards."""
	
	def __init__(self, name, color='', type_=None, types='', subtypes='', text='', power='', toughness='', cmc='0', cost='', loyalty='', isCopy=False, isToken=True, matchOnName=None):
		self.name = titlecase(name)
		self.type_ = type_
		self.color = color
		if self.color != 'Pink' and self.color != 'Gold':
			self.color = ''.join(orderColors(color))
		self.cmc = cmc
		self.cost = cost
		if len(text) >= 2:
			text = text[0].upper() + text[1:] #Capitalizes the first letter
		if 'Cumulative upkeep' in text or 'Bands with other' in text:
			text = text[:-1] #Strip the erroneous period.
		self.text = text
		self.power = str(power)
		if self.power == 'X':
			self.power = '*'
		self.toughness = str(toughness)
		if self.toughness == 'X':
			self.toughness = '*'
		self.isCopy = isCopy
		#Is the name alone enough to consider this a match?
		#For emblems, embalm tokens, and things that shouldn't
		#have naming collisions.
		self.matchOnName = matchOnName
		self.isToken = isToken
		self.loyalty = loyalty
		
		if self.isToken == 'quasi' and self.matchOnName is None:
			self.matchOnName = True
		
		if type_ is not None:
			#Use the supplied entire type line.
			self.type_ = type_
		else:
			#Types and subtypes were supplied separately.
			if not 'creature' in types.casefold() and self.power != '':
				types = ' '.join(filter(None, (types, 'Creature')))
			if not subtypes and 'creature' in types.casefold():
				subtypes = self.name
			self.type_ = ' — '.join(filter(None, (types, subtypes)))
		#Capitalize.
		self.type_ = self.type_.title()
		#Prepend "token" if it's not already there.
		if not 'Token' in self.type_:
				self.type_ = ' '.join(filter(None, ('Token', self.type_)))
	
	@classmethod
	def copyCard(cls, card):
		return cls(
			name=card.name,
			type_=card.type_, #Automatically prepends "Token"
			color=card.color,
			cmc=card.cmc,
			cost=card.cost,
			text=card.text,
			power=card.power,
			toughness=card.toughness,
			loyalty=card.loyalty,
			isCopy=True,
			isToken=True,
		)
	
	def __repr__(self):
		return f"<TempToken: '{self.lackeyName}'>"
	
	def __hash__(self):
		#Note that this is not actually immutable!
		#But we shouldn't be changing any of these fields once
		#the token is recorded into a record.
		attributes = (
			self.name,
			self.color,
			self.type_,
			self.text,
			self.power,
			self.toughness,
		)
		return hash(attributes)
	
	def __eq__(self, other):
		if self.color == 'WUBRG':
			#Five-color tokens (so far) print this like rules text.
			wubrg = r'(?:This creature is all colors\.)|'
		else:
			wubrg = ''
		try:
			if all([
				self.name == other.name,
				self.color == other.color,
				self.type_ == other.type_,
				#Note: Scryfall's API doesn't give curly quotes or apostrophes. So I
				#can't differntiate an apostrophe from a single quote in quoted rules text.
				re.sub(r'[\s,.\'"]', '', self.text).casefold() == re.sub(wubrg + r'[\s,.\'"]', '', other.text).casefold(),
				self.power == other.power,
				self.toughness == other.toughness,
			]):
				return True
			else:
				return False
		except AttributeError:
			return False
	
	def asTuple(self):
		"""Returns an immutable tuple that can be used as a key."""
		return (
			self.name,
			self.type_,
			#Cost and CMC are only used for self-copies, which shouldn't ever need this.
			self.color,
			self.power,
			self.toughness,
			re.sub(r'\s', '+', self.text).casefold()
		)
	
	@property
	def lackeyName(self):
		if self.isToken:
			if self.isToken == 'quasi':
				#Backside or flip, not an actual token
				return f'[XXX] {self.name}'
			elif self.isCopy:
				return f'[XXX] {self.name} (Copy)'
			else:
				outputName = self.name
				parenthetical = []
				if self.color != '':
					parenthetical.append(self.color)
				power = self.power
				toughness = self.toughness
				if power != '':
					if power == 'X':
						power = '*'
					if toughness == 'X':
						toughness = '*'
					parenthetical.append(f'{power}/{toughness}')
					
					commaItems = []
					if 'Enchantment' in self.type_ and 'Creature' in self.type_:
						commaItems.append('Enchantment')
					
					keywords = re.findall(r'^([^.]+?)\n?$', self.text, re.MULTILINE) #If there are any lines with no periods, it's just keywords
					if keywords:
						keywords = tuple('Bands with Other' if 'bands with other'.casefold() in keyword.casefold() else keyword for keyword in keywords)
						commaItems += keywords
						#parenthetical.append(', '.join(keywordsMatches).title())
					
					if commaItems:
						parenthetical.append(titlecase(', '.join(commaItems)))
				if parenthetical:
					parenthetical = ' '.join(parenthetical)
					outputName = f'{self.name} ({parenthetical})'
				
				return f'[XXX] {outputName}'
		
		return self.name

class TokenRerouteRule(metaclass=Recorded):
	""""
	Saves a rule for which set a token should spawn from if its token
	wasn't printed in its set.
	
	For now, stubbed out and calling the old method internally.
	"""
	key = 'tokenTuple' #Immutable so it can be a trustworthy key.
	
	def __init__(self, token, defaultSet, specificReroutes=[]):
		"""
		Can take a token or a token tuple; always stores a tuple.
		Can take an actual set object or its code string; always stores the string, for better serializing
		Specific reroutes should be SpecificReroute objects.
		Not fully implemented yet
		"""
		if type(token) == TempToken:
			self.tokenTuple = token.asTuple()
		elif type(token) == tuple:
			self.tokenTuple = token
		else:
			raise ValueError("TokenRerouteRule's token must be either a TempToken or a tuple.")
		
		if type(defaultSet) == Set:
			self.defaultSetCode = defaultSet.code
		elif type(defaultSet) == str:
			self.defaultSetCode = defaultSet
		else:
			raise ValueError("TokenRerouteRule's default set must be either a Set or a string code.") 
		
		#Not implemented yet:
		#self.specificReroutes = specificReroutes
	
	def displaySet(self):
		set_ = Set[self.defaultSetCode]
		return f'{set_.name} ({set_.code})'
	
	@classmethod
	def load(cls, filename='tokenRerouteRules'):
		"""Load saved rules from file."""
		try:
			with open(f"../data/{filename}.json", "r") as readFile:
				ruleDicts = json.load(readFile)
				for ruleDict in ruleDicts:
					cls.fromDict(ruleDict)
		except FileNotFoundError:
			raise FileNotFoundError('No file of saved reroute rules found.')
	
	@classmethod
	def fromDict(cls, dict_):
		"""Construct a TokenRerouteObject from a dict that has been read in from stored JSON."""
		return cls(tuple(dict_['tokenTuple']), dict_['defaultSetCode'])
	
	@classmethod
	def findForToken(cls, token):
		"""
		Find an existing reroute rule for the specified token, or return None.
		"""
		try:
			if type(token) == TempToken:
				rule = cls[token.asTuple()]
			elif type(token) == tuple:
				rule = cls[token]
			else:
				raise ValueError('Can only find a rule for a token or token tuple.')
			#Reroute rule already exists.
			return rule
		except KeyError:
			#No existing rule for this token.
			return None
	
	@indented
	def replaceTokeninScript(self, script):
		"""
		Replaces the TempToken in a script with the token (Card object) from the corrected set.
		For now, only implements default sets, not specific reroutes.
		"""
		if script.token.asTuple() != self.tokenTuple:
			raise ValueError(f"Tried to make {self} replace {script.token}, but it doesn't match")
		
		newToken = Set[self.defaultSetCode].fetchToken(script.token, noHasteOkay=True)
		if newToken:
			display(3, f'Replacing {script.token} in {script.set_} with {newToken}')
			script.token = newToken
		else:
			raise KeyError(f"Tried to replace {script.token} in {script.set_} with token from {Set[self.defaultSetCode]}, but none found")
	
	def __repr__(self):
		return f"<TokenReroute Rule for {self.tokenTuple}>"

class MissedTokenRecord(metaclass=Recorded):
	"""
	Keeps track of tokens that don't exist in their current set/block,
	and don't already have a saved reroute rule.
	
	Stores tokens as tuples for indexing, but can store actual sets,
	since this won't be serialized.
	"""
	key = 'tokenTuple'
	
	def __init__(self, token, createImage=True):
		self.tokenTuple = token.asTuple()
		self.setsThatNeed = []
		self.setsThatHave = []
		self.customToken = None
		self.noHasteMatch = False
		#'createImage' parameter means 'generate an image if this ends up making a custom token.
		
		#Look for any matching tokens. Here's where it GETS REAL.
		display(3, f"{token} hasn't been missed before. Searching the plugin for it!")
		
		#Suppress the million messages we'll get from this.
		savedVerbosity = verbosity_setting.verbosity
		verbosity_setting.verbosity = 0
		
		if not token.isCopy:
			for set_ in Set.instancesList():
				foundToken = set_.fetchToken(token)
				if foundToken:
					print(f'Found match in {set_}')
					self.setsThatHave.append(set_)
		
		if len(self.setsThatHave) == 0 and token.text == 'Haste':
			for set_ in Set.instancesList():
				foundToken = set_.fetchToken(token, noHasteOkay=True)
				if foundToken:
					print(f'Found no-haste match in {set_}')
					self.noHasteMatch = True
					self.setsThatHave.append(set_)
		
		#Restore verbosity.
		verbosity_setting.verbosity = savedVerbosity
		
		if len(self.setsThatHave) == 1 and not TokenRerouteRule.findForToken(token):
			TokenRerouteRule(token, self.setsThatHave[0])
		
		if not self.setsThatHave:
			#CREATE A NEW FUCKING CAAAAAARD YEEEAAAAH
			display(3, 'Creating new Lackey token: {token}')
			if not createImage:
				print('This is a playtest card! Make a token without an image.')
			lackey = Set['lackey']
			self.customToken = Card.fromToken(token, lackey, image=createImage)
			lackey.cards.append(self.customToken)
			if not TokenRerouteRule.findForToken(token):
				TokenRerouteRule(token, lackey)
			
			#If this isn't a copy, check and see if it exists on Scryfall
			if not token.isCopy:
				color = token.color
				if not color:
					color = 'c'
				query = [
					f'!"{token.name}"',
					'is:token',
					f'power={token.power}',
					f'toughness={token.toughness}',
					f'type:"{token.type_}"',
					f'color={color}',
				]
				if token.text:
					query.append('oracle:/^' + token.text.replace('\n', r'\n') + '$/')
				else:
					query.append('is:vanilla')
				query = ' '.join(query)
				if scryfallRequest(params={'q':query}):
					click.launch(f'https://scryfall.com/search?q={query}&unique=prints')
	
	@classmethod
	def findForToken(cls, token):
		"""
		Return a record for the specified token
		if it exists, or return None.
		"""
		try:
			if type(token) == TempToken:
				record = cls[token.asTuple()]
			elif type(token) == tuple:
				record = cls[token]
			else:
				raise ValueError('Can only find a rule for a token or token tuple.')
			return record
		except KeyError:
			return None
	
	@classmethod
	def findOrCreate(cls, token, createImage=True):
		"""Return a record, either retrieved or new."""
		try:
			return cls[token.asTuple()]
		except KeyError:
			return MissedTokenRecord(token, createImage=createImage)
	
	@classmethod
	def reviewTokens(cls, saveFile='tokenRerouteRules'):
		display(1, 'Tokens not found in their own set or block:')
		Indent.increase()
		for record in cls.instancesList():
			display(1, record.tokenTuple)
			#Need to check if they have only one set, or no set, in which case don't ask about it.
			if record.customToken:
				display(1, '\tUsing a custom Lackey token.')
				continue
			if len(record.setsThatHave) == 1:
				set_ = record.setsThatHave[0]
				display(1, f'\tUsing only available set: {set_.name} ({set_.code})')
				if record.noHasteMatch:
					display(1, '\t\t***(A no-haste match)')
				continue
			
			display(1, 'Sets that need it:')
			Indent.increase()
			for set_ in record.setsThatNeed:
				display(1, f'{set_.name} ({set_.code})')
			Indent.decrease()
		
			display(1, 'Sets that have it:')
			Indent.increase()
			if record.noHasteMatch:
				display(1, '***(Except without haste)')
			for set_ in record.setsThatHave:
				display(1, f'{set_.name} ({set_.code})')
			Indent.decrease()
		
			rule = TokenRerouteRule.findForToken(record.tokenTuple)
			defaultMsg = ''
			default = None
			if rule:
				display(1, f'Current reroute rule: {rule.displaySet()})')
				defaultMsg = f'. (Default: leave current setting, {rule.defaultSetCode})'
				default = rule.defaultSetCode
			else:
				display(1, 'No existing reroute rule found for this token.')
		
			needInput = True
			while needInput:
				response = click.prompt("Enter code to set new rule, or 'view' to view tokens on Scryfall" + defaultMsg, default=default)
				if rule and response == rule.defaultSetCode:
					#Leave everything the same.
					print('Leaving setting intact.')
					needInput = False
				elif response in [set_.code for set_ in record.setsThatHave]:
					#Change or create the rule.
					print(f'Setting rule to {response}')
					if rule:
						rule.defaultSet = response
					else:
						TokenRerouteRule(record.tokenTuple, response)
					needInput = False
				elif response == 'view':
					print('Opening Scryfall...')
					name, type_, color, power, toughness, text = record.tokenTuple
					name = f'!"{name}"'
					type_ = re.sub(r'Token\s?', '', type_)
					if type_:
						type_ = f't:"{type_}"'
					if color == '':
						color = 'c'
					color = f'c={color}'
					if power:
						power = f'power:{power}'
					if toughness:
						toughness = f'toughness:{toughness}'
					if text:
						text = text.replace('/', '\/')
						text = f'o:/{text}/'
					sets = '+or+'.join([f'e:t{set_.code}' for set_ in record.setsThatHave])
					constraints = '+'.join(filter(None, (name, type_, color, power, toughness, text)))
					click.launch(f'https://scryfall.com/search?unique=prints&q={constraints}+({sets})')
		
		Indent.decrease()
		
		#Go back and resolve any applicable scripts.
		for category in Category.instancesList():
			category.resolveSpawnScripts(review=False)
		
		#Save the generated reroute rules to a file:
		with contextlib.suppress(FileNotFoundError):
			os.remove(f'../data/{saveFile}.json')
		with open(f"../data/{saveFile}.json", "w") as writeFile:
		   json.dump([rule.__dict__ for rule in TokenRerouteRule.instancesList()], writeFile, indent='\t', ensure_ascii=False)
	
	def register(self, set_):
		if not set_ in self.setsThatNeed:
			self.setsThatNeed.append(set_)
		#if not self.setsThatHave:
		#	pass

class LackeyOriginRecord(metaclass=Recorded):
	"""Records the origin card(s) of each custom Lackey token."""
	
	key = 'myTokenTuple'
	
	def __init__(self, token):
		self.token = token
		self.originCards = []
	
	@property
	def myTokenTuple(self):
		return self.token.asTuple()
	
	@classmethod
	def register(cls, token, originCard):
		"""Called each time a card makes a Lackey token."""
		try:
			record = cls[token.asTuple()]
		except KeyError:
			record = cls(token)
		
		record.originCards.append(originCard)
	
	def uniqueCardsForJson(self):
		"""Gets the latest (in Lackey, not always chronologically) print for each card, excluding promos."""
		uniqueCards = {}
		for card in self.originCards:
			if card.promoSetCode is None:
				uniqueCards[card.name] = card
		
		returnList = []
		for card in uniqueCards.values():
			cardDict = {
				'name': escape(card.name),
				'imageFile': escape(card.imageFile),
			}
			returnList.append(cardDict)
		return returnList
	
	@classmethod
	def saveToFile(cls):
		#Save a JSON record of all the Lackey tokens, and which card(s) create each:
		with contextlib.suppress(FileNotFoundError):
			os.remove(f'../data/lackey_origins.json')
		
		recordsDict = {}
		for record in LackeyOriginRecord.instancesList():
			recordsDict[record.token.lackeyName] = record.uniqueCardsForJson()
		
		with open(f"../data/lackey_origins.json", "w") as writeFile:
		   json.dump(recordsDict, writeFile, indent='\t', ensure_ascii=False)

class Format(metaclass=Recorded):
	"""A format for the selector.
	Stores set objects.
	"""
	def __init__(self, name, setsList):
		"""Takes a a list containing sets
		and/or containers that contain sets"""
		self.name = name
		self.sets = []
		for item in setsList:
			if hasattr(item, 'sets'):
				self.sets += [set_ for set_ in item.sets]
			elif hasattr(item, 'cards'):
				self.sets.append(item)
			else:
				raise AttributeError(f"Can't add {item} to {self.name} format, because it contains neither sets nor cards.")
	
	@classmethod
	def fromYAML(cls, formatKey, formatValue):
		setsList = []
		for item in formatValue:
			setsList.append(fetchContainer(item))
		return cls(formatKey, setsList)
	
	def __repr__(self):
		return f"<Format '{self.name}'>"
	
	def output(self):
		"""
		Returns a string of how this format should be saved to file.
		Has to be named differently and assigned at instantiation. (I wrote that sentence, but what does it mean?)
		"""
		output = f'<format><label>{self.name}</label>\n'
		for set_ in self.sets:
			output += f'\t<set>{set_.lackeyCode}</set>\n'
			if set_.promos:
				for promoCode in set_.promos:
					output += f'\t<set>{promoCode}</set>\n'
		output += '</format>\n\n'
		return output
	
	@classmethod
	def writeFormats(cls):
		"""
		Writes formats file
		"""
		fileName = os.path.join(updateFilesDir, 'formats.txt')
		os.makedirs(os.path.dirname(fileName), exist_ok=True)
		display(1, f'Writing {fileName}...')
		with open(fileName, 'w', encoding='ascii') as f:
			f.write('<formatdefinitions>\n\n')
			for format in cls.instancesList():
				f.write(format.output())
			f.write('</formatdefinitions>\n')

class PackCategory(metaclass=Recorded):
	"""A category of packs, each with a title pack"""
	numberOfPackFiles = 4
	
	def __init__(self, name, packsList):
		self.name = name
		self.packs = [TitlePack(name)]
		for item in packsList:
			if hasattr(item, 'packs'):
				self.packs += reversed([pack for pack in item.packs])
			elif isinstance(item, Pack):
				self.packs.append(item)
			else:
				raise ValueError(f"Can't add {item} to {self.name} pack category, because it has no packs.")
		#Remove recent sets:
		if self.name != 'Recent Sets':
			filtered_packs = []
			for pack in self.packs:
				if pack not in PackCategory['Recent Sets'].packs:
					filtered_packs.append(pack)
			self.packs = filtered_packs
	
	@classmethod
	def fromYAML(cls, packKey, packValue):
		contentsList = []
		for item in packValue:
			if isinstance(item, dict):
				#It's a declaration for an individual pack.
				classLookup = {
					'Pack': Pack,
					'Booster': Booster,
					'SeededBooster': SeededBooster,
				}
				className = item.get('type', 'Booster') #Default to Booster
				item.pop('type', None) #Delete the type setting if it's there
				
				if item.get('set'):
					item['set_'] = Set[item['set']] #Replace the code string with the actual set
					item.pop('set')
				
				if item.get('structure'):
					item['structure'] = NamedPackStructure[item['structure']].structure
				
				if item.get('commonCrits'):
					item['commonCrits'] = [CommonCrit(*crit) for crit in item['commonCrits']]
				pack = classLookup[className](**item) #Call the constructor
				contentsList.append(pack)
			else:
				#It's a category or set.
				contentsList.append(fetchContainer(item))
		return cls(packKey, contentsList)
	
	@classmethod
	def writePackFiles(cls):
		"""Write all the pack files, splitting into the specified number."""
		packList = [pack for category in PackCategory.instancesList() for pack in category.packs]
		display(3, f'Pack list:')
		Indent.increase()
		for pack in packList:
			display(3, pack)
		Indent.decrease()
		
		#Split up the list into n - 1 chunks. (The last is copied from a static template.)
		chunkedPackList = numpy.array_split(packList, cls.numberOfPackFiles - 1)
		
		display(1, 'Writing pack files')
		os.makedirs(os.path.join(updateFilesDir, 'packs'), exist_ok=True)
		
		for index, chunk in enumerate(chunkedPackList):
			fileName = os.path.join(updateFilesDir, 'packs', f'packdefinitions{index + 1}.xml')
			display(2, f'\tWriting {fileName}')
			with open(fileName, 'w') as f:
				f.write('<packdefinitions version="1.0">\n\n')
				f.write(''.join(pack.xml for pack in chunk))
				f.write('</packdefinitions>')
		
		#I don't need to recreate the random M12 decks in Python, but I want to include them.
		
		display(2, f'\tWriting file packdefinitions4.xml from template')
		shutil.copy('../templates/m12packdefinitions.xml', f'{updateFilesDir}/packs/packdefinitions4.xml')
