#!/usr/bin/env python
#Functions for comparing plugin versions
#Run in the "lackey-python3" virtualenv.

import os
import re
import pathlib
from string import Template
from datetime import datetime
import html
import shutil
import pathlib
from slugify import slugify
from datetime import datetime
import verbosity_setting
from classes import display, Indent

def buildCardListFromFiles(cardDir):
	#Reads existing card files and builds a list from them, for making comparisons
	keys = 	()
	cards = {}
	for fileName in [f for f in os.listdir(cardDir) if '.txt' in f]:
		with open(cardDir + '/' + fileName) as f:
			firstLine = True
			#Skip the first line, since it's just headers.
			for line in f:
				if firstLine:
					keys = line.replace('\n', '').split('\t')[2:] #The first two are always name and set.
					firstLine = False
					continue
				if not line.isspace():
					values = line.replace('\n','').split('\t')
					key = '{} ({})'.format(values[0], values[1]) #name and set
					values = dict(zip(keys, values[2:])) #everything else
					cards[key] = values
	return cards

def buildPackListFromFiles(packDir):
	#Reads existing pack files and builds a list from them, for making comparisons
	if os.path.exists(packDir):
		packs = {}
		matches = []
		for fileName in [f for f in os.listdir(packDir) if 'packdefinitions' in f]:
			with open(packDir + '/' + fileName) as f:
				contents = f.read()
				matches += re.findall(r'<pack>\n\t<packtitle>(.+?)</packtitle>(.+?)</pack>', contents, re.DOTALL)
		if matches:
			return {match[0]:match[1] for match in matches}
		else:
			return {}
	else:
		return {}

def getVersionMessage(baseDir):
	path = baseDir +  '/version.txt'
	if os.path.exists(path):
		with open(path, 'r') as f:
			contents = f.read()
			return re.search(r'<message>(.+)</message>', contents).groups()[0]
	else:
		return '[No log message]'

def cardChangeTable(changedCards):
	#Used by compare to generate its file. Replace with a template?
	if changedCards:
		output = '\t<table>\n'
		for cardName, changes in changedCards.items():
			output += f'\t\t<tr><td class="name" colspan="3">{cardName}</td></tr>\n'
			for key, change in changes.items():
				output += f'\t\t<tr class="changes card {key}"><td>{key}</td><td>{html.escape(change[0])}</td><td>{html.escape(change[1])}</td></tr>\n'
				output += '\t\t<tr class="spacer"><td></td></tr>\n'
		output += '\t</table>\n'
		return output
	else:
		return ''

def packChangeTable(changedPacks):
	#Used by compare to generate its file. Replace with a template?
	if changedPacks:
		output = '\t<table>\n'
		for packName, texts in changedPacks.items():
			output += f'\t\t<tr><td class="name" colspan="2">{packName}</td></tr>\n'
			output += f'\t\t<tr class="changes pack"><td>{html.escape(texts[0])}</td><td>{html.escape(texts[1])}</td></tr>\n'
			output += '\t\t<tr class="spacer"><td></td></tr>\n'
		output += '\t</table>\n'
		return output
	else:
		return ''

def simpleList(title, listOfCards):
	#Used by compare to generate its file. Replace with a template?
	if listOfCards:
		output = '\t<h2>{}</h2>\n'.format(title)
		if len(listOfCards):
			output += '\t<ul>\n'
			for cardName in listOfCards:
				output += '\t\t<li>{}</li>\n'.format(cardName)
			output += '\t</ul>\n'
		else:
			output += '\t<p>None</p>\n'
		return output
	else:
		return ''

class Section():
	def __init__(self, title, count, contents):
		self.title = title
		self.slug = slugify(title)
		self.count = count
		self.nonzero = ''
		if self.count > 0:
			self.nonzero = ' nonzero'
		self.contents = contents
		self.navActive = ''
		self.divActive = ''
	
	def activate(self):
		self.navActive = ' class="active"'
		self.divActive = 'active '

def getVersion(age):
	#Gets the directory and message of update files for age specified.
	if age == None:
		raise ValueError("Can't get version without specifying an age.")
	baseDir = '../generatedFiles/'
	backups =  [folder for folder in os.listdir(baseDir) if 'backup' in folder]
	backups.sort(reverse=True)
	versions = ['update'] + backups
	versionDir = versions[age]
	return {'dir': versionDir, 'message': getVersionMessage(baseDir + versionDir)}

def compare(fileName='../data/changes.html', old=1, new=0):
	#Compare two sets of generated card files. By default, compares current against most recent backup.
	#Change "old" and/or "new" parameters to compare other pairs of versions.
	display(1, 'Comparing versions for changes')
	if not old >= new:
		raise ValueError('Old directory must be older (higher number) than new directory.')
	
	baseDir = '../generatedFiles'
	
	oldVersion = getVersion(old)
	oldDir = oldVersion['dir']
	oldMessage = oldVersion['message']
	
	newVersion = getVersion(new)
	newDir = newVersion['dir']
	newMessage = newVersion['message']
	
	interveningLogs = ''
	if old - new > 1:
		numInBetween = old - new - 1
		interveningLogs = f'<h2>Intervening versions: {numInBetween}</h2>\n\n'
		#Gather intervening logs, if they exist.
		for i in reversed(range(new+1, old)):
			message = getVersion(i)['message']
			interveningLogs += f'<p class="message">{message}</p>\n'
	
	oldCards = buildCardListFromFiles('/'.join((baseDir, oldDir, 'cardFiles')))
	newCards = buildCardListFromFiles('/'.join((baseDir, newDir, 'cardFiles')))
	
	display(2, 'Old:')
	Indent.increase()
	display(2, f'Number: {old}')
	display(2, f'Directory: {oldDir}')
	display(2, f'Log message: {oldMessage}')
	Indent.decrease()
	
	display(2, 'New:')
	Indent.increase()
	display(2, f'Number: {new}')
	display(2, f'Directory: {newDir}')
	display(2, f'Log message: {newMessage}')
	Indent.decrease()
	
	#Check for fields that were added or removed.
	removedFields = [i for i in list(oldCards.values())[0].keys() if not i in list(newCards.values())[0].keys()]
	addedFields = [i for i in list(newCards.values())[0].keys() if not i in list(oldCards.values())[0].keys()]
	
	display(2, 'Checking cards...')
	Indent.increase()
	#List comprehensions aren't quite as pretty as sets, but if we want
	#consistent order, it's silly to convert to sets and back.
	removedCards = [i for i in oldCards if not i in newCards]
	addedCards = [i for i in newCards if not i in oldCards]
	maintainedCards = [i for i in newCards if i in oldCards]
	changedCards = {}
	
	for cardName in removedCards:
		display(3, f'\tRemoved: {cardName}')
	for cardName in addedCards:
		display(3, f'\tAdded: {cardName}')
	#Compare the cards that stayed, and see if any of their fields changed.
	for cardName in maintainedCards:
		#print(cardName)
		oldCard = oldCards[cardName]
		newCard = newCards[cardName]
		changes = {}
		for key in oldCard:
			if newCard.get(key, None) is not None: #Make sure it's still a key in the newer file!
				if oldCard[key] != newCard[key]:
					display(3, f'\tChanged: {cardName}: {key}')
					#print('\tOld: "{}"'.format(oldCard[key]))
					#print('\tNew: "{}"'.format(newCard[key]))
					changes[key] = (oldCard[key], newCard[key])
					if len(changes):
						#One or more things changes, so save this "pair" of cards to the change list.
						changedCards[cardName] = changes
	
	Indent.decrease()
	#Check packs for changes too.
	display(2, 'Checking packs...')
	Indent.increase()
	oldPacks = buildPackListFromFiles('/'.join((baseDir, oldDir, 'packs')))
	newPacks = buildPackListFromFiles('/'.join((baseDir, newDir, 'packs')))
	
	removedPacks = [i for i in oldPacks if not i in newPacks]
	addedPacks = [i for i in newPacks if not i in oldPacks]
	maintainedPacks = [i for i in newPacks if i in oldPacks]
	changedPacks = {}
	
	for packName in removedPacks:
		display(3, f'\tRemoved: {packName}')
	for packName in addedPacks:
		display(3, f'\tAdded: {packName}')
	
	for packName in maintainedPacks:
		oldText = oldPacks[packName]
		newText = newPacks[packName]
		if oldText != newText:
			display(3, f'\tChanged: {packName}')
			changedPacks[packName] = (oldText, newText)
	Indent.decrease()
	
	sections = (
		Section('Added/Removed Fields', len(removedFields) + len(addedFields), simpleList('Removed Fields', removedFields) + simpleList('Added Fields', addedFields)),
		Section('Added/Removed Cards', len(removedCards) + len(addedCards), simpleList('Removed Cards', removedCards) + simpleList('Added Cards', addedCards)),
		Section('Changed Cards', len(changedCards), cardChangeTable(changedCards)),
		Section('Added/Removed Packs', len(removedPacks) + len(addedPacks), simpleList('Removed Packs', removedPacks) + simpleList('Added Packs', addedPacks)),
		Section('Changed Packs', len(changedPacks), packChangeTable(changedPacks)),
	)
	
	#If at least one section has changed, set the earliest one as active.
	#Otherwise, set the first section as active.
	unchanged = True
	for section in sections:
		if section.count > 0:
			section.activate()
			unchanged = False
			break
	if unchanged:
		sections[0].activate()
	
	navTabs = ''
	sectionDivs = ''
	
	for section in sections:
		navTabs += f'\t\t<li><a href="#"{section.navActive} onClick="activate(this, \'{section.slug}\'); return false;">{section.title} <span class="count{section.nonzero}">{section.count}</span></a></li>\n'
		sectionDivs += f'<div id="{section.slug}" class="{section.divActive}section">\n{section.contents}</div>\n\n'
	
	values = {
		'oldDir': oldDir,
		'old': old,
		'oldMessage': oldMessage,
		'interveningLogs': interveningLogs,
		'newDir': newDir,
		'new': new,
		'newMessage': newMessage,
		'navTabs': navTabs,
		'sectionDivs': sectionDivs,
	}
	with open('../templates/changes-template.html') as f:
		template = Template(f.read())
	
	display(1, f'Writing changes file: {fileName}')
	with open(fileName, 'w') as f:
		f.write(template.substitute(values))
