import os
import regex as re
from string import Template
from datetime import datetime
from contextlib import nullcontext
import requests
import requests_cache
from PIL import Image

requests_cache.install_cache('scryfall_cache', allowable_codes=(200, 404))

updateFilesDir = '../generatedFiles/update'
imagesDir = '../images'
versionUrl = 'http://magicplugin.normalitycomics.com/update/version.txt'
updateUrl = 'http://magicplugin.normalitycomics.com/update/updatelist.txt'

import verbosity_setting

#Helpers and general use stuff

class Indent():
	"""
	Handle adjustabe levels of indentation.
	"""
	indentation = 0
	
	@classmethod
	def tabs(cls):
		return '\t' * cls.indentation
	
	@classmethod
	def increase(cls, increment=1):
		cls.indentation += increment
	
	@classmethod
	def decrease(cls, increment=1):
		cls.indentation -= increment
		if cls.indentation < 0:
			raise ValueError('Somehow reached negative indentation')

def display(level, *messages, sep='\n'):
	"""
	Print or don't print, according to verbosity level,
	and at the current indentation
	"""
	if verbosity_setting.verbosity >= level:
		output = []
		for message in messages:
			output.append(Indent.tabs() + str(message).replace('\n', '\n' + Indent.tabs()))
		print(sep.join(output))

def indented(func):
	""""
	Indent the decorated function's display by one level.
	For some reason, breaks initializers... dunno why.
	"""
	def returnFunc(*args, **kwargs):
		#print(args)
		#print(kwargs)
		Indent.increase()
		returnValue = func(*args, **kwargs)
		Indent.decrease()
		return returnValue
	
	return returnFunc

class MultiDict(dict):
	"Dictionary that can assign multiple keys. Can't use tuples as keys, though."
	def __setitem__(self, key, value):
		if type(key) is tuple:
			for subkey in key:
				super().__setitem__(subkey, value)
		else:
			super().__setitem__(key, value)

class Recorded(type):
	"""
	Any class inheriting from this metaclass can get
	its instances using their indices as dictionary keys.
	Also enforces unique indices.
	By default, the keyed field is 'name', but this can
	be overidden in the class by defining the 'key' class variable.
	"""
	
	#These three methods get assigned to the class that's instantiated.
	def _get(cls, key, default=None):
		return cls._instances.get(key, default)
	
	
	def _record(cls, instance):
		keyOfInstance = getattr(instance, instance.__class__.key)
		#print(f'Recording {cls} as key {keyOfInstance}')
		if cls.get(keyOfInstance):
			raise NameError(f'Already a {cls.__name__} at key {keyOfInstance}')
		cls._instances[keyOfInstance] = instance
	
	
	def _instancesList(cls):
		"""
		Return a list that's a COPY of the dictionary's values.
		Should be fine as long as the source doesn't change
		during iteration. Used so that the order can be reversed.
		"""
		return list(cls._instances.values())
	
	
	#This one just STAYS in the metaclass, since that's where Python looks for it.
	#Using this to make index notation work is the entire reason I'm
	#using a metaclass instead of ordinary inheritance.
	def __getitem__(cls, key):
		try:
			return cls._instances[key]
		except KeyError:
			raise KeyError(f'No {cls.__name__} at key {key}')
	
	
	def __init__(cls, name, bases, dct):
		""""
		The initializer for the CLASS
		Give the class an empty instances dict,
		as well as its class methods.
		"""
		cls._instances = {}
		if not hasattr(cls, 'key'):
			cls.key = 'name'
		cls.record = classmethod(cls.__class__._record)
		cls.get = classmethod(cls.__class__._get)
		cls.instancesList = classmethod(cls.__class__._instancesList)
		super().__init__(name, bases, dct)
	
	
	def __call__(cls, *args, **kwargs):
		"""
		What happens when you call the instantiated class,
		i.e. when you create an instance. Delegate to
		the class's defined initializer, but record
		the new instance in that class's instance list.
		"""
		instance = type.__call__(cls, *args, **kwargs)
		cls.record(instance)
		return instance

@indented
def scryfallRequest(params, cached=True, root='https://api.scryfall.com/', method='cards/search'):
	"""
	Handle not only sending the request, but dealing with 
	Scryfall's pagination. Return a list of searched cards.
	Note: each item returned is Scryfall's JSON representation
	of a card, not a Card object. This means each DFC will be returned
	as one card, with two faces.
	"""
	
	# display(1,
	# 	f'params: {params}\n\ttype: {type(params)}',
	# 	f'root: {root}\n\ttype: {type(root)}',
	# 	f'method: {method}\n\ttype: {type(method)}',
	# 	f'cached: {cached}\n\ttype: {type(cached)}',		
	# 	sep='\n'
	# 	)
	
	#Scryfall doesn't always like the default percent encodings.
	#Giving Requests a string instead of a dictionary for params
	#overrides its auto-escaping.
	params = '&'.join(f'{key}={value}' for key, value in params.items())
	params = params.replace(' ', '+')
	
	if cached:
		context = nullcontext()
	else:
		context = requests_cache.disabled()
	
	with context:
		response = requests.get(root + method, params=params)
		display(2, f'Requesting URL: {response.url}')
		page = response.json()
	
		allPages = [page]
	
		if not page.get('data', False):
			display(2, 'No cards returned')
			return []
	
		while page['has_more']:
			page = requests.get(page['next_page']).json()
			allPages.append(page)
	
	jsonCards = []
	
	#Excluded card names:
	excludeList = (
		'Punchcard',
		'Checklist'
	)
	
	for page in allPages:
		for jsonCard in page['data']:
			#Exclude certain cards we don't need.
			exclude = False
			for i in excludeList:
				if i in jsonCard['name']:
					exclude = True
			if not exclude:
				jsonCards.append(jsonCard)
	
	display(2, f'Cards fetched: {len(jsonCards)}')
	return jsonCards

@indented
def getChecksum(fileName):
	"""Reimplementation of Lackey's checksum function.
	It doesn't work right on image files, but that's
	okay, I'm only running it on text files.
	"""
	checksum = 0
	with open(fileName, 'rb') as f:
		byte = f.read(1)
		while byte:
			n = int.from_bytes(byte, byteorder='big')
			if n != 10 and n != 13:
				checksum = checksum + n
			checksum %= 100000000
			byte = f.read(1)
		checksum -= 1 #In C++, EOF returns -1
		checksum %= 100000000
	display(4, f'Checksum for {fileName}: {checksum}')
	return checksum

def escape(text):
	"Escapes all the non-ASCII characters we're likely to find, plus newlines."
	replacements = {
		'“': '"',
		'”': '"',
		'’': "'",
		'½': '1/2',
		'∞': '{Infinity}',
		'π': '{Pi}',
		'²': '^2',
		'—': '-',
		'−': '-',
		'•': '-',
		'☐': '-',
		'®': '',
		'á': 'a',
		'â': 'a',
		'â': 'a',
		'à': 'a',
		'é': 'e',
		'í': 'i',
		'ö': 'o',
		'ú': 'u',
		'û': 'u',
		'★': 's',
		'\n//\n': ' // ',
		'\n': ' | ',
	}
	
	for r in replacements.items():
		text = text.replace(*r)
	
	#Checks if any remaining characters are non-ASCII, and prints what it was.
	try:
		text.encode('ascii')
	except UnicodeEncodeError as e:
		char = re.search(r"character ('.+?')",str(e)).group(1)
		print('\nFound non-ASCII character ({}) in: \n{}\n'.format(char, text))
		raise
	return text

def orderColors(colors):
	"""Sort a list of colors into consistent WUBRG order."""
	colorOrders = {
		2: [
			['W', 'U'],
			['W', 'B'],
			['R', 'W'],
			['G', 'W'],
			['U', 'B'],
			['U', 'R'],
			['G', 'U'],
			['B', 'R'],
			['B', 'G'],
			['R', 'G'],
		],
		3: [
			['W', 'U', 'B'],
			['U', 'B', 'R'],
			['B', 'R', 'G'],
			['R', 'G', 'W'],
			['G', 'W', 'U'],
			['W', 'B', 'G'],
			['U', 'R', 'W'],
			['B', 'G', 'U'],
			['R', 'W', 'B'],
			['G', 'U', 'R'],
		],
		4: [
			['W', 'U', 'B', 'R'],
			['U', 'B', 'R', 'G'],
			['B', 'R', 'G', 'W'],
			['R', 'G', 'W', 'U'],
			['G', 'W', 'U', 'B'],
		],
	}
	
	length = len(colors)
	if length <= 1:
		return colors
	elif length != 5:
		for colorOrder in colorOrders[length]:
			if sorted(colors) == sorted(colorOrder):
				return colorOrder
	else:
		return ['W', 'U', 'B', 'R', 'G']

def convertColors(colorNames):
	"""
	Convert a list of color names to letter abbreivations
	(and order them)
	"""
	mapping = {
		"white": 'W',
		'blue': 'U',
		'black': 'B',
		'red': 'R',
		'green': 'G',
	}
	colors = []
	for i in colorNames:
		colors.append(mapping[i])
	return orderColors(colors)

def extractColors(colorString):
	"""
	Given a string like 'white, blue, and black', return a string of
	the color abbreviation letters, in WUBRG order
	"""
	if 'gold' in colorString:
		return 'Gold'
	if 'pink' in colorString:
		return 'Pink'
	
	pColors = '|'.join(('white', 'blue', 'black', 'red', 'green', 'colorless'))
	colors = re.findall(pColors, colorString)
	if 'colorless' in colors or 'gold' in colors:
		return ''
	return ''.join(convertColors(colors))

def titlecase(string, delimiter=' '):
	"""Return a string in something approximating correct title case (for token names, including keywords lists)"""
	input_words = string.lower().split(delimiter)
	output_words = []
	small_words = ['the', 'a', 'an', 'with','from', 'of', 'to', 'by', 'as', 'at', 'and']
	first_word = True
	for word in input_words:
		if re.match(r'\{.+\}$', word): #Card symbols
			output_words.append(word.upper())
		elif '-' in word:
			output_words.append(titlecase(word, delimiter='-'))
		elif word in small_words and not first_word:
			output_words.append(word.lower())
		else:
			output_words.append(word.capitalize())
			first_word = False
	return delimiter.join(output_words)

def cnRange(query, setCode, cached):
	"""Convert collector number comparison search terms in a query to an "or" list of numbers."""
	terms = query.replace('+', '').split()
	newTerms = []
	minimum = None
	maximum = None
	
	for term in terms:
		match = re.match(r'(?:cn|number)(?P<comparison><|>)(?P<equals>=)?(?P<number>\d+)$', term, re.IGNORECASE)
		if match:
			if match.group('comparison') == '>':
				if minimum is not None:
					raise ValueError("Query contains multiple minimum collector numbers.")
				minimum = int(match.group('number'))
				if match.group('equals') is None:
					minimum += 1
			if match.group('comparison') == '<':
				if maximum is not None:
					raise ValueError("Query contains multiple maximum collector numbers.")
				maximum = int(match.group('number'))
				if match.group('equals') is None:
					maximum -= 1
		else:
			newTerms.append(term)
					
	if minimum is None and maximum is None:
		return ' '.join(terms)
	elif minimum is None:
		minimum = 1
	elif maximum is None:
		params = {
			'order': 'set',
			'dir': 'desc',
			'unique': 'prints',
			'q': f'e:{setCode}'
		}
		savedVerbosity = verbosity_setting.verbosity
		verbosity_setting.verbosity = 0
		maximum = int(scryfallRequest(params, cached=cached)[0]['collector_number'])
		verbosity_setting.verbosity = savedVerbosity
	
	if minimum >= maximum:
		raise ValueError("Query specifies a minimum that's higher than its specififed maximum.")
	
	parenthetical = '(' + ' or '.join(f'cn:{number}' for number in range(minimum, maximum+1)) + ')'
	newTerms.append(parenthetical)
	return ' '.join(newTerms)
