#!/usr/bin/env python
# Multi-purpose command line tool for updating the Lackey plugin

#Standard
import contextlib
import os
import regex as re
import pysftp
import html
import datetime
import json
#Installed
import click
import keyring
import numpy
#Mine
import verbosity_setting
import mb1

#Still using things from here, for now...
import comparison

from classes import *
from tokenimaging import TokenImage

@click.group()
def cli():
	'''Multi-purpose command line tool for updating the Lackey plugin.
	Invoke a command with --help for details on its usage.'''
	
	pass

@cli.command() #cli, not click! The instance we made above.
@click.option('--images/--no-images', default=False, help='Download ALL images? (Default=false, only download sets listed as "new" in settings file)')
@click.option('--review-tokens/--fast-tokens', default=False, help='Go over token reroute rules (Default=false, simply use saved rules for token spawns)')
@click.option('-v', '--verbosity', default=1, help='How much detail: 1–3 (default=2, lists sets but not cards)')
def import_all(images, verbosity, review_tokens):
	"""Imports all Magic sets specified in the settings.
	Also generates all the files necessary
	for updating, and checks changes."""
	verbosity_setting.verbosity = verbosity
	updateMessage = loadConfig()
	
	if images:
		for set_ in Set.instancesList():
			set_.getImages = True
	
	for category in Category.instancesList():
		category.getCards()
	
	#Import the saved reroute rules:
	TokenRerouteRule.load()
	
	for category in Category.instancesList():
		category.resolveSpawnScripts(review=review_tokens)
	
	if review_tokens:
		MissedTokenRecord.reviewTokens()
	
	namesSeen = []
	namesWithDuplicates = []
	
	for card in Set['lackey'].cards:
		if card.lackeyName in namesSeen:
			namesWithDuplicates.append(card.lackeyName)
		else:
			namesSeen.append(card.lackeyName)
	
	if namesWithDuplicates:
		print('Custom token names with duplicates:')
		for name in namesWithDuplicates:
			print('\n' + name)
	else:
		print('No duplicate custom token names found.')
	
	mb1.add_draft_qualities()
	
	LackeyOriginRecord.saveToFile()
	
	# doubleSpawnCards = []
	# 	for category in Category.instancesList():
	# 		for card in category.cards:
	# 			if not card.name in doubleSpawnCards: 
	# 				tokens = []
	# 				for script in card.callableScripts:
	# 					if script.token:
	# 						if not script.token in tokens:
	# 							tokens.append(script.token)
	# 				if len(tokens) > 1:
	# 					doubleSpawnCards.append(card.name)
	# 	names = ' or '.join([f'!"{name}"' for name in doubleSpawnCards])
	# 	click.launch(f'https://scryfall.com/search?unique=cards&q={names}')
	
	# cardsWithNoSpawns = []
	# for set_ in Set.instancesList():
	# 	for card in set_.cards:
	# 		if 'create'.casefold() in card.text.casefold():
	# 			if not card.spawnScripts:
	# 				cardsWithNoSpawns.append(card.name)
	# query = '+or+'.join(f"%21%22{name.replace(' ', '+')}%22" for name in cardsWithNoSpawns)
	# click.launch('https://scryfall.com/search?q=' + query)
	
	#Delete the generated data files, if they exist.
	print('\nDeleting previous update files folder.')
	with contextlib.suppress(FileNotFoundError):
		shutil.rmtree(updateFilesDir)
	
	for category in Category.instancesList():
		category.writeCardFile()
	
	print('Checking Mystery Booster cards:')
	mb1.check_for_cards()
	
	Category.writeListOfCardDataFiles()
	
	PackCategory.writePackFiles()
	
	Format.writeFormats()
	writeVersionFile(updateMessage)
	writeUpdateList()
	#This should probably live elsewhere eventually.
	comparison.compare()

@cli.command()
@click.option('--from-settings/--from-dirs', default=True, help="Which sets to upload. Default is to upload sets listed as downloading images in settings. From-dirs uploads all images in images folder.")
def upload_images(from_settings):
	"""Upload images for new sets."""
	verbosity_setting.verbosity = 1
	
	if from_settings:
		print('Checking settings file...')
		imageSets = [set_ for set_ in Set.instancesList() if set_.new]
		if len(imageSets) == 0:
			display(1, 'No new sets defined in settings file.')
			return
		display(1, 'New sets defined in settings file:')
		Indent.increase()
		
		missingTotal = 0
		setCodes = []
		for set_ in imageSets:
			display(1, f'{set_.name} ({set_.code})')
			Indent.increase()
			if os.path.isdir(os.path.join(imagesDir, set_.code)):
				display(1, 'Found image folder')
				setCodes.append(set_.code)
			else:
				display(1, 'Missing image folder!!!')
				missingTotal += 1
			if os.path.isdir(os.path.join(imagesDir, 't' + set_.code)):
				display(1, f'Found token image folder: t{set_.code}')
				setCodes.append('t' + set_.code)
			else:
				display(1, 'Warning: no token image folder')
			Indent.decrease()
		Indent.decrease()
		display(1, '')
		if missingTotal > 0:
			display(1, f'{missingTotal} set(s) without matching image directory. ABORT!')
			return
	else:
		display(1, 'Checking images directory...')
		setCodes = []
		for item in os.scandir(imagesDir):
			if item.is_dir():
				setCodes.append(item.name)
		if len(setCodes) == 0:
			display(1, 'No set folders contained in images folder.')
			return
	
	#Either way, now we have a list of actual set folders to upload.
	display(1, 'Sets to be uploaded:', *setCodes, sep='\n\t')
	display(1, '')
	
	click.confirm("Okay to upload the listed image folders? Will overwrite existing images, but won't delete any others in existing folders.", abort=True)
	
	password = keyring.get_password('magicplugin', 'normycgn')
	with pysftp.Connection(host='198.54.114.227', username='normycgn', port=21098, password=password) as sftp:
		for setcode in setCodes:	
			display(1, f'Uploading {setcode}...')
			Indent.increase()
			try:
				sftp.mkdir(f'magicplugin.normalitycomics.com/cardimages/{setcode}')
				display(1, 'Creating remote image folder')
				display(1, 'Copying into it...')
			except IOError:
				display(1, 'Remote image folder exists. Copying into it...')
			sftp.put_r(f'../images/{setcode}', f'magicplugin.normalitycomics.com/cardimages/{setcode}')
			Indent.decrease()

@cli.command()
@click.option('--age', default=0, help='How old of a version to upload. (default=0, current update folder)')
def upload_files(age):
	"""Upload newest data files."""
	password = keyring.get_password('magicplugin', 'normycgn')
	version = comparison.getVersion(age)
	print(f'Using version aged {age}:\n\tLog Message: {version["message"]}')
	if age != 0:
		print('CAUTION: THIS IS NOT THE LATEST VERSION!')
	
	click.confirm("This will upload the directory, replacing what's on the server. Okay to continue?", abort=True)
	
	with pysftp.Connection(host='198.54.114.227', username='normycgn', port=21098, password=password) as sftp:
		print('Uploading directory...')
		sftp.put_r('../generatedFiles/update', 'magicplugin.normalitycomics.com/update')

@cli.command()
@click.argument('setcode')
@click.option('--custom-query', default=None, type=click.STRING, help="Custom API query. By default, use the one in settings. Specify 'blank' to explicitly omit.")
@click.option('--images/--no-images', default=False, help='Download image files? (default=off)')
@click.option('--include-tokens/--no-tokens', default=True, help='Include tokens? (default=on)')
@click.option('--review-tokens/--fast-tokens', default=False, help='Go over token reroute rules (Default=false, simply use saved rules for token spawns)')
@click.option('-v', '--verbosity', default=3, help='How much detail: 1–3 (default=3, everything)')
def import_set(setcode, custom_query, images, include_tokens, verbosity, review_tokens):
	"""Imports one set (for debugging, usually)."""
	verbosity_setting.verbosity = verbosity
	loadConfig()
	
	#Reimplement options.
	set_ = Set[setcode]
	if images:
		set_.getImages = True
	category = Category(setcode, [set_])
	category.getCards()
	
	TokenRerouteRule.load()
	
	category.resolveSpawnScripts(review=review_tokens)
	
	if review_tokens:
		MissedTokenRecord.reviewTokens()
	
	LackeyOriginRecord.saveToFile()
	
	category.writeCardFile(destDir='../sets')
	

@cli.command()
@click.option('-n', '--name', 'specified_name', default=None, help='Only save a specific token by name.')
@click.option('-t', '--type', 'specified_type', default=None, help='Only save tokens of the specified type.')
@click.option('--only-copies/--all', default=False, help='Only save tokens that are copies.')
@click.option('--hi-res/--low-res', default=False, help='Save tokens as uncropped full-res, or as scaled-down and cropped versions for Lackey.')
@click.option('--jace/--not-jace', default=False, help="Save only a high-res, artless image of Jace's frame for compositing.")
def save_token_images(specified_name, specified_type, only_copies, jace, hi_res):
	"""Save images for custom Lackey tokens, as last saved to file"""
	
	loadConfig()
	
	verbosity_setting.verbosity = 3
	
	with contextlib.suppress(FileNotFoundError):
		os.remove('../data/token_lightbox_links.txt')
	
	with open(f"../data/lackey_origins.json", "r") as readFile:
		originsDict = json.load(readFile)
	
	with open(updateFilesDir + '/cardFiles/custom-tokens-for-lackeyccg.txt') as f:
		firstLine = True
		#Skip the first line, since it's just headers.
		for line in f:
			if firstLine:
				firstLine = False
				continue
			if not line.isspace():
				values = line.replace('\n','').split('\t')
				(
					name,
					set_,
					imageFile,
					color,
					colorId,
					cost,
					cmc,
					type_,
					power,
					toughness,
					loyalty,
					rarity,
					draftQualities,
					sound,
					script,
					text
				) = values
				if imageFile == '':
					continue
				name = name.replace('[lackey] ', '')
				if name[-7:] == ' (Copy)':
					isCopy = True
				else:
					isCopy = False
				name = re.sub(' \(.*?\)', '', name)
				card = Card(
					name=name,
					set_=Set['lackey'],
					color=color,
					cost=cost,
					cmc=cmc,
					rarity=rarity,
					type_=type_.replace(' - ',' — '),
					power=power,
					toughness=toughness,
					text=text.replace(' | ', '\n').replace(' - ', ' — '),
					isToken=True,
					isCopy=isCopy
				)
				
				web = True
				if web:
					token_root = '../cardimages/'
					card_root = '../cardimages'
				else:
					token_root = '../../images'
					card_root = 'http://magicplugin.normalitycomics.com/cardimages'
				
				originCards = originsDict[card.lackeyName.replace('lackey', 'XXX')]
				
				originImages = [f'\t\t<img id="{originCard["imageFile"].replace("/", "")}" src="{card_root}/{originCard["imageFile"]}.jpg" class="mouse-hovered">\n' for originCard in originCards]
				originIds = ' '.join([originCard['imageFile'].replace('/', '') for originCard in originCards])
				
				originNames = [escape(originCard['name']).replace(' ', '+') for originCard in originCards]
				query = '+or+'.join([f'!&quot;{originName}&quot;' for originName in originNames])
				
				with open('../data/token_lightbox_links.txt', 'a') as f:
					f.write(''.join(originImages))
					f.write(f'\t\t<a class="hover" data-image-id="{originIds}" href="https://scryfall.com/search?q={query}&amp;unique=cards">\n')
					f.write(f'\t\t\t<img src="{token_root}/{card.tokenImageFileName()}.jpg" alt="{card.lackeyName[9:]}"/>\n')
					f.write('\t\t</a>\n')
				if jace:
					if name == 'Jace, Cunning Castaway':
						click.confirm("Delete existing custom Jace image?", abort=True)
						with contextlib.suppress(FileNotFoundError):
							os.remove('../templates/tokens/custom/jace-cunning-castaway-copy.jpg')
						card.text = '...'
						jace = TokenImage(card, type_line_shift=-3)
						jace.art_layer = None
						background, title_line, text_box, type_line, set_symbol, brush = jace.frame_layers
						jace.frame_layers = [background, title_line, set_symbol, brush]
						jace.save((745, 1040), crop=False, fileName='../PSDs/Jace Frame.png')
						break
					else:
						continue
				if specified_name is not None and specified_name.casefold() != name.casefold():
					continue
				if only_copies and not card.isCopy:
					continue
				if specified_type is not None and not card.isType(specified_type):
					continue
				image = TokenImage(card)
				if hi_res:
					image.save(crop=False, size=image.template_size, forcepng=True)
				else:
					image.save()

@cli.command()
@click.argument('setcode')
def check_partials(setcode):
	"""Open Scryfall to check if partially spoiled cards have been updated with better images."""
	verbosity_setting.verbosity = 0
	print(f'Checking partial images for updates in {Set[setcode].name} ({setcode})')
	partials = Set[setcode].partials
	parenthetical = '+or+'.join([f'number:{number}' for number in partials])
	query = f'set:{setcode}+({parenthetical})'
	url = f'https://scryfall.com/search?unique=cards&order=set&q={query}'
	
	print(f'\nOpening partials:\n{url}')
	click.launch(url)
	
	#Get names of all the cards currently listed as partials, so we can exclude those.
	params = {
		'order': 'set',
		'unique': 'cards',
		'q': query,
	}
	namesOfPartials = [jsonCard['name'].replace('"', '') for jsonCard in scryfallRequest(params, cached=False)]
	excludes = '+'.join(f'-!"{name}"' for name in namesOfPartials)
	query = f'set:{setcode}+{excludes}'
	url = f'https://scryfall.com/search?order=set&unique=cards&q={query}'
	
	print(f'\nOpening complete cards:\n{url}')
	click.launch(url)

@cli.command()
def check_custom_tokens():
	"""Check to see if any of the currently defined custom tokens have been actually printed."""
	with open(updateFilesDir + '/cardFiles/custom-tokens-for-lackeyccg.txt') as f:
		firstLine = True
		#Skip the first line, since it's just headers.
		for line in f:
			if firstLine:
				firstLine = False
				continue
			if not line.isspace():
				values = line.replace('\n','').split('\t')
				(
					name,
					set_,
					imageFile,
					color,
					colorId,
					cost,
					cmc,
					type_,
					power,
					toughness,
					loyalty,
					rarity,
					draftQualities,
					sound,
					script,
					text
				) = values
				name = name.replace('[lackey] ', '')
				if name[-7:] == ' (Copy)' or ('Ooze' in name and 'Gutter Grime' in text):
					continue
				display(1, f'Checking for matches for {name}...')
				name = re.sub(' \(.*?\)', '', name)
				#Note: currently won't support cards with quotation marks in them, but non-copy
				#custom tokens haven't yet had those.
				if text:
					text = re.sub(r'\s', '+', text).lower()
					text_term = f'o:"^{text}$"'
				else:
					text_term = 'is:vanilla'
				if color == '':
					color = 'c'
				#types = [each_type for each_type in type_.split() if each_type != '-' and each_type != 'Token']
				#type_terms =  ' '.join(f't:{each_type}' for each_type in types)
				type_ = type_.replace('-', '—')
				query = f'!"{name}" c:{color} type:"{type_}" power:{power} toughness:{toughness} {text_term} is:token'
				params = {
					'order': 'set',
					'unique': 'prints',
					'q': query
				}
				verbosity_setting.verbosity = 0
				jsonCards = scryfallRequest(params, cached=False)
				verbosity_setting.verbosity = 1
				
				Indent.increase()
				if jsonCards:
					display(1, f'Found {len(jsonCards)} matches. Opening...')
					url = f'https://scryfall.com/search?unique=prints&order=set&q={query}'
					click.launch(url)
					click.pause()
				else:
					display(1, 'No matches found.')
				Indent.decrease()

@cli.command()
@click.option('-o', '--old', default=1, help='How old a version to check against (default=1, most recent backup)')
@click.option('-n', '--new', default=0, help='How old a version to check (default=0, current)')
@click.option('-v', '--verbosity', default=3, help='How much detail: 1–3 (default=3, everything)')
def compare(old, new, verbosity):
	"""Compares versions of the plugin."""
	verbosity_setting.verbosity = verbosity
	comparison.compare(old=old, new=new)

@cli.command()
@click.option('-v', '--verbosity', default=2, help='How much detail: 1–3 (default=2')
def packs(verbosity):
	"""Generates pack files."""
	loadConfig()
	verbosity_setting.verbosity = verbosity
	PackCategory.writePackFiles()

@cli.command()
@click.option('-m', '--message', default=None, help='Update message to use. Default: use from file')
@click.option('-v', '--verbosity', default=3, help='How much detail: 1–3 (default=3, everything')
def change_message(verbosity, message):
	"""Generate an updatelist file for the current version."""
	verbosity_setting.verbosity = verbosity
	if message is None:
		message = loadConfig() # also loads the categories
	writeVersionFile(message)
	writeUpdateList()
	comparison.compare()

@cli.command()
@click.option('-v', '--verbosity', default=3, help='How much detail: 1–3 (default=3, everything')
def updatelist(verbosity):
	"""Generate an updatelist file for the current version."""
	verbosity_setting.verbosity = verbosity
	writeUpdateList()

@cli.command()
@click.argument('fileName')
def checksum(filename):
	"""Get the checksum of a specified file (only works on text files, not images)."""
	getChecksum(filename)

@cli.command()
def backup():
	"""Back up the most recently generated update files."""
	oldName = '../generatedFiles/update'
	newName = '../generatedFiles/backup ' + str(datetime.now())[:19].replace(':', '.')
	print('Backing up update files')
	print(f'Copying {oldName} to {newName}')
	shutil.copytree(oldName, newName)

#To make this run as a standalone:
if __name__ == '__main__':
    cli()