#!/usr/bin/env python
#Library to generate pack definitions for Lackey
#Created by CrazyChucky. Instructions at http://www.lackeyccg.com/forum/index.php?topic=12713.0
#For use with LackeyCCG (http://lackeyccg.com/) which is © Trevor Agnitti
#Written using Python 3.7

#SETTINGS

#List the fields your plugin supports. This lets you catch
#typos if you try to check against a nonexistant field.
possibleFields = (
	'Name',
	'Set',
	'ImageFile',
	'Color',
	'ColorID',
	'Cost',
	'ConvertedCost',
	'Type',
	'Power',
	'Toughness',
	'Rarity',
	'DraftQualities',
	'Script',
	'Text',
)

#Evaluations that Lackey supports. Supposedly there's an
#ISGREATERTHAN, but in testing it appears not to function.
possibleEvals = (
	'IS',
	'CONTAINS',
	'DOESNTCONTAIN',
	#'ISGREATERTHAN',
)

#Lackey doesn't currently support probabilities with a decimal
#component. If that ever changes, this library can generate
#more accurate probabilities by changing this flag to False.
roundToIntegers = True

from collections import Counter

tab = '    '

#These are the classes (and one function) you'll likely be using:

class Pack:
	"""Represents a single pack"""
	def __init__(self, name, packKind, structure=None, chances=None, commonCrits=None):
		self.name = name
		self.packKind = packKind
		#Build a branchpoint cointaining a branch containing the supplied structure, resolve it,
		#and get the resulting branches. They're now fully resolved into Lackey's pack-wide "chances".
		if chances is not None:
			self.chances = chances
		elif structure is not None:
			self.chances = resolveToChances(structure)
		else:
			raise ValueError('Must supply either structure or chances to Pack constructor.')
		self.commonCrits = commonCrits
	
	@property
	def xml(self):
		"""Output the pack's XML."""
		xml = f'<pack>\n\t<packtitle>{self.name}</packtitle>\n\t<packkind>{self.packKind}</packkind>\n'
		if self.commonCrits:
			xml += ''.join([commonCrit.xml for commonCrit in self.commonCrits])
		xml += ''.join([chance.xml for chance in self.chances])
		xml += '</pack>\n\n'
		
		return xml
	
	def __repr__(self):
		return f'<{self.__class__.__name__}: {self.name}>'

class BranchPoint:
	"""
	Represents a decision-point with in a pack (for example, whether we have a mythic or a rare).
	All of a branchpoint's branches' probabilities should sum to 100%.
	"""
	def __init__(self, *branches):
		self.branches = tuple(branches) #Nice to be immutable
		probs = [branch.prob for branch in self.branches]
		totalProb = sum(probs)
		if totalProb != 100:
			stringProbs = ' + '.join([str(prob) for prob in probs])
			raise ValueError(f'Tried to create branch with total probability of {totalProb}% ({stringProbs})')
	
	def display(self, indent=0):
		tabs = indent * tab
		return f'{tabs}<BranchPoint:\n' + '\n'.join(branch.display(indent+1) for branch in self.branches) + f'\n{tabs}>'
		
	def __repr__(self):
		#Don't try to fully represent this for normal purposes, you'll go blind.
		strProbs = ', '.join([str(branch.prob) + '%' for branch in self.branches])
		return f'<BranchPoint: {len(self.branches)} Branches ({strProbs})>'
	
	def __eq__(self, other):
		if hasattr(other, 'branches'):
			return self.branches == other.branches
		else:
			return False

def combineLikeKinds(kinds):
	"""Recursively combine any combinable kinds in this branch."""
	kinds = list(kinds)
	for firstIndex, firstKind in enumerate(kinds):
		for secondIndex, secondKind in enumerate(kinds[firstIndex+1:], start=firstIndex+1):
			if firstKind.isCombinable(secondKind):
				combinedKind = firstKind + secondKind
				kinds = kinds[:firstIndex] + [combinedKind] + kinds[firstIndex+1:secondIndex] + kinds[secondIndex+1:]
				return combineLikeKinds(kinds)
	
	return tuple(kinds)

class Branch:
	"""
	Each of the possibilities within a branchpoint, and how likely each one is.
	A branch's entries can consist of kinds and/or more branchpoints.
	Lackey's "chance" XML tag corresponds to a branch containing no branchpoints.
	"""
	def __init__(self, *entries, prob=100):
		
		#Combine any combinable kinds. Only worth doing if we're otherwise fully resolved.
		if all(isinstance(entry, Kind) for entry in entries):
			entries = combineLikeKinds(entries)
		self.entries = entries
		
		self.prob = prob
		if self.prob < 0:
			raise ValueError("Can't create a branch with negative probability. This might happen if you try to subdivide probabilities into\
			tiny pieces and rounding is turned on.")
	
	@property
	def xml(self):
		if all([not hasattr(entry, 'branches') for entry in self.entries]):
			#If none of the entries have branches, this is a pure "chance" and should have XML.
			contents = ''.join([kind.xml for kind in self.entries])
			return f'\t<chance>\n\t\t<prob>{self.prob}</prob>\n{contents}\t</chance>\n'
		else:
			raise ValueError(f"{self} contains sub-branches and thus can't output XML.")
	
	def display(self, indent=0):
		tabs = indent * tab
		#return f'{tabs}<Branch: ' + '\n'.join(tabs + entry.display(indent+1) for entry in self.entries) + '{tabs}>'
		if self.prob == 100:
			return '\n'.join(f'{tabs}{entry.display(indent)}' for entry in self.entries)
		
		oneLiner =  f'{tabs}{self}'
		if len(oneLiner) <= 180 and not any(isinstance(entry, BranchPoint) for entry in self.entries):
			return oneLiner
		
		return f'{tabs}<Branch:\n' + '\n'.join(entry.display(indent+1) for entry in self.entries) + f'\n{tabs}  Prob: {self.prob}%' + f'\n{tabs}>'

	def __repr__(self):
		strEntries = ', '.join([str(entry) for entry in self.entries])
		return f'<Branch: prob={self.prob}%, entries=[{strEntries}]>'
	
	def __eq__(self, other):
		if hasattr(other, 'entries') and hasattr(other, 'prob'):
			return compareEntries(self.entries, other.entries) and (self.prob == other.prob)
		else:
			return False
	
	def __add__(self, other):
		if hasattr(other, 'entries') and hasattr(other, 'prob'):
			if compareEntries(self.entries, other.entries):
				return Branch(*self.entries, prob=self.prob + other.prob)
			else:
				raise ValueError("Can't combine two branches with unequal sets of entries.")
		else:
			raise TypeError("Can't add a branch and " + str(type(other)))

class Kind:
	"""Corresponds to Lackey's "kind" XML tag. A given number of cards, within specified criteria"""
	def __init__(self, qty, *crits):
		self.qty = qty
		self.crits = crits
		contents = f'\n\t\t\t<qty>{self.qty}</qty>\n'
		contents += ''.join([f'\t\t\t{crit.xml}\n' for crit in self.crits]) + '\t\t'
		self.xml = f'\t\t<kind>{contents}</kind>\n'
	
	def display(self, indent=0):
		tabs = indent * tab
		return f'{tabs}{self}'
	
	def isCombinable(self, other):
		return Counter([str(crit) for crit in self.crits]) == Counter([str(crit) for crit in other.crits])
	
	def __add__(self, other):
		return self.__class__(self.qty + other.qty, *self.crits)
	
	def __repr__(self):
		strCrits = ', '.join(str(crit) for crit in self.crits)
		return f'<Kind: {self.qty} of {strCrits}>'

class Crit():	
	"""Corresponds to Lackey's "crit" (criterion) XML tag. Specifies requirements for a kind."""
	def __init__(self, field, critEval, data, reversed=False):
		#I should pull this out to some sort of global setting,
		#for other people to more easily use this module.
		
		if field in possibleFields:
			self.field = field
		else:
			raise ValueError(f'{field} is not a valid field for crit.')
		if critEval in possibleEvals:
			self.eval = critEval
		else:
			raise ValueError(f'{critEval} is not a valid eval for crit.')
		self.data = data
	
	@property
	def xml(self):
		return f'<crit><field>{self.field}</field><eval>{self.eval}</eval><data>{self.data}</data></crit>'	
	
	def display(self, indent=0):
		tabs = indent * tab
		return f'{tabs}{self}'
	
	def __repr__(self):
		return f'<Crit: {self.field} {self.eval} {self.data}>'
	
	def reverse(self):
		if self.eval == 'CONTAINS':
			return self.__class__(self.field, 'DOESNTCONTAIN', self.data)
		elif self.eval == 'DOESNTCONTAIN':
			return self.__class__(self.field, 'CONTAINS', self.data)
		else:
			raise ValueError(f"'{self.eval}' does not have a defined opposite.")

class CommonCrit(Crit):
	"""
	Corresponds to Lackey's "commoncrit" XML tag. Same as regular crits, but apply to a whole pack.
	The only difference here is the class name and  XML output.
	"""
	@property
	def xml(self):
		#Don't just replace the word, in case a card's name or something includes "crit".
		return f'\t<packcommoncrit><field>{self.field}</field><eval>{self.eval}</eval><data>{self.data}</data></packcommoncrit>\n'
	
	def __repr__(self):
		return f'<CommonCrit: {self.field} {self.eval} {self.data}>'

def writePacksToFile(packs, fileName):
	#Take a collection of packs and write them to the specified file.
	with open(fileName, 'w') as f:
		f.write('<packdefinitions version="1.0">\n\n')
		f.write(''.join(pack.xml for pack in packs))
		f.write('</packdefinitions>')


#Below are functions that deal with the internal nuts and bolts of resolving branches.
#Feel free to check them out and mess with them! But you shouldn't need to call them directly
#for just defining packs.

def resolveToChances(structure):
	if not isinstance(structure, Branch):
		structure = Branch(*structure)
	return combineLikeBranches(resolveBranchPoint(BranchPoint(structure)).branches)

def resolveBranchPoint(branchPoint):
	#Recursively digs through a branchpoint and splits up any branchpoints in it,
	#into the pack-wide probabilities required by Lackey.
	newBranches = []
	for branch in branchPoint.branches:
		newBranches += resolveBranch(branch) #concatenate, since it always returns a list
	newBranchPoint = BranchPoint(*newBranches)
	if branchPoint != newBranchPoint:
		#There's been a change, so run again.
		branchPoint = newBranchPoint
		return resolveBranchPoint(newBranchPoint)
	else:
		return newBranchPoint

def resolveBranch(branch):
	#Split up a branch if it contains a branch point. (Split at the first one encountered.)
	#Always returns a list, even if it's just the one unchanged branch.
	firstBranchPoint = False
	startingProb = branch.prob
	for index, entry in enumerate(branch.entries): #Save the index so we know where the bp is in the list of entries.
		if hasattr(entry, 'branches'):
			firstBranchPoint = entry
			break
	if firstBranchPoint:
		#We want to make a new set of branches to replace the one we're inspecting,
		#with their probabilities as a proportion of the main branch's starting
		#probabilitiy.
		newBranches = []
		for childBranch in firstBranchPoint.branches:
			#Split up the allocated probability.
			precision = 2
			newProb = round((childBranch.prob * startingProb / 100), precision) #Round to avoid repeating or long decimals.
			if roundToIntegers:
				newProb = int(round(newProb, 0))
				if newProb == 0:
					newProb = 1
			#Each new branch should have the branchpoint's spot replaced by a resolved option from it.
			newEntries = branch.entries[:index] + childBranch.entries + branch.entries[index+1:]
			newBranch = Branch(*newEntries, prob=newProb)
			newBranches.append(newBranch)
		#We're not quite done! We should make sure the (possibly rounded) probabilities equal 100%.
		#This really only comes up with decimal probabilities, which Lackey doesn't support (yet) anyway, but this is ready for them.
		total = sum([branch.prob for branch in newBranches])
		difference = round(startingProb - total, precision)
		if difference != 0:
			#The (sort of hacky) fix is to just adjust the last entry to make it fit.
			lastBranch = newBranches[-1]
			newBranches[-1] = Branch(*lastBranch.entries, prob=lastBranch.prob + difference)
		return newBranches
	else:
		#If there were no branchpoints, return the branch as is, but in a list.
		return [branch]

def combineLikeBranches(branches):
	#Sometimes, after resolving all the branchpoints, we may be left with
	#multiple branches containing the same entries. These should be combined
	#to be tidier.
	#Only run this on fully resolved sets of branches (containing no branchpoints).
	branches = list(branches)
	for index, branch in enumerate(branches):
		for matchIndex, matchBranch in enumerate(branches[index+1:], start=index+1):
			if compareEntries(branch.entries, matchBranch.entries):
				combinedBranch = branch + matchBranch
				#Replace the first branch with the combination, and remove its match.
				newBranches = branches[:index] + [combinedBranch] + branches[index+1:matchIndex] + branches[matchIndex+1:]
				#Run again on the result to find more matches, if any.
				return combineLikeBranches(newBranches)
	#If no matches are found, return the original branches unchanged.
	return branches	

def compareEntries(a, b):
	#See if two sets of entries are the same, regardless of order.
	#Return True if so
	return Counter([str(entry) for entry in a]) == Counter([str(entry) for entry in b])


