#!/usr/bin/env python
#Magic-specific code for generating Lackey packs
from lackeyPacks import *
from helpers import Recorded, display, Indent
from ruamel.yaml import YAML
from ruamel.yaml.constructor import SafeConstructor

from pprint import pprint
#Load the crits and structure config file.

def loadPacksConfig():
	"""Load YAML file and create named pack structures, which can be accessed via key."""
	yaml = YAML(typ='safe')
	
	def construct_yaml_map(self, node):
		# test if there are duplicate node keys
		firstKey = self.construct_object(node.value[0][0], deep=True)
		if isinstance(firstKey, str):
			#Create a dictionary, as normal.
			data = {}
		else:
			#If the key isn't a string, it's a probability. Create a list instead.
			data = []
		yield data
		for key_node, value_node in node.value:
			key = self.construct_object(key_node, deep=True)
			val = self.construct_object(value_node, deep=True)
			if isinstance(data, dict):
				data[key] = val
			else:
				data.append((key, val))

	SafeConstructor.add_constructor(u'tag:yaml.org,2002:map', construct_yaml_map)
	modded_yaml = YAML(typ='safe')
	
	with open('config/packs.yaml') as yamlFile:
	#with open('config/packs-backup.yaml') as yamlFile:
		#documents = yaml.load_all(yamlFile)
		documents = modded_yaml.load_all(yamlFile)
		critsData, structuresData = list(documents)

	for key, value in critsData.items():
		RecordedCrit(key, *value)
	for key, value in structuresData.items():
		NamedPackStructure(key, value).structure
	
	with open('config/seededPacks.yaml') as yamlFile:
		seededData = yaml.load(yamlFile)
	
	for key, value in seededData.items():
		SeededPackRecord(key, value)
	
class NamedPackStructure(metaclass=Recorded):
	"""Represents a saved pack's tructure, keyed to name."""
	key = 'key'
	
	def __init__(self, name, structure):
		#Use hyphens to assign multiple keys to the same NamedPackStructure object.
		self.name = name
		if '-' in name:
			keys = name.split('-')
			self.key = keys[0]
			for otherKey in keys[1:]:
				self.__class__._instances[otherKey] = self
		else:
			self.key = name
		
		if isinstance(structure, list):
			#self.structure = nodeFromYAMLBackup(structure)
			self.structure = nodeFromYAML(structure)
		else:
			self.structure = structure
	
	def __repr__(self):
		return f'<Named pack structure: {self.name}>'

class RecordedCrit(Crit, metaclass=Recorded):
	def __init__(self, name, field, critEval, data, reversed=False):
		self.name = name
		super().__init__(field, critEval, data)
		#If this crit is reversible (and isn't already reversed), also create its opposite.
		if not reversed:
			if self.eval == 'CONTAINS':
				notEval = 'DOESNTCONTAINS'
			elif self.eval == 'DOESNTCONTAIN':
				notEaval = 'CONTAINS'
			else:
				return
			newName = f'not{self.name.capitalize()}'
			self.__class__(newName, self.field, 'DOESNTCONTAIN', self.data, reversed=True)

def nodeFromYAML(node):
	"""Recursively parse a pack structure, supplied as a YAML-derived list, into the appopriate Python objects."""
	if isinstance(node, tuple):
		#A branch
		if isinstance(node[0], int) and node[1] == []:
			#A branch with no kinds
			return Branch(prob=node[0])
		elif isinstance(node[1][0], int):
			#A branch containing a single kind
			return Branch(nodeFromYAML(node[1]), prob=node[0]) 
		else:
			#A branch with a list of kinds
			return Branch(*nodeFromYAML(subnode for subnode in node[1]), prob=node[0]) 
	elif isinstance(node, list) and isinstance(node[0], int):
		#A kind
		qty, *critSpecs = node
		crits = []
		for critSpec in critSpecs:
			if isinstance(critSpec, str):
				crit = RecordedCrit[critSpec]
			else:
				crit = Crit(*critSpec)
			crits.append(crit)
		return Kind(qty, *crits)
	else:
		#Investigate all subnodes.
		results = [nodeFromYAML(subnode) for subnode in node]
		if all(isinstance(result, Branch) for result in results):
			#A branchpoint: contains only branches 
			return BranchPoint(*results)
		else:
			#The fully resolved list of entries
			return results

def nodeFromYAMLBackup(node):
	"""Recursively parse a pack structure, supplied as a YAML-derived list, into the appopriate Python objects."""
	if (
		isinstance(node, list) 
		and node[0] == []
		and isinstance(node[1], int)
		and len(node) == 2
	):
		#A branch with no kinds
		return Branch(prob=node[1])
	elif isinstance(node, int):
		#A probilitiy
		return node
	elif isinstance(node[0], int):
		#A kind
		qty, *critNames = node
		crits = [RecordedCrit[critName] for critName in critNames]
		return Kind(qty, *crits)
	else:
		#Investigate all subnodes.
		results = [nodeFromYAMLBackup(subnode) for subnode in node]
		if (
			all(isinstance(result, Kind) or isinstance(result, BranchPoint) for result in results[:-1])
			and isinstance(results[-1], int)
		):
			#A branch: contains only kinds and/or branchpoints, plus a final probability
			return Branch(*results[:-1], prob=results[-1])
		elif all(isinstance(result, Branch) for result in results):
			#A branchpoint: contains only branches 
			return BranchPoint(*results)
		else:
			#The fully resolved list of entries
			return results

class Booster(Pack):
	#Default booster pack for Magic. Automatically inserts a commonCrit for the
	#set code. Specify more if necessary.
	def __init__(self, set_, structure=None, name=None, commonCrits=None):
		allCommonCrits = [CommonCrit('Type', 'DOESNTCONTAIN', 'Token'), CommonCrit('DraftQualities', 'DOESNTCONTAIN', 'notInPacks')]
		if set_.code != 'tsp':
			allCommonCrits = [CommonCrit('Set', 'IS', set_.lackeyCode)] + allCommonCrits
		
		#Should be covered by "notInPacks" already. Leaving here in case I need it later.
		#if set_.promos:
		#	allCommonCrits.append(CommonCrit('DraftQualities', 'DOESNTCONTAIN', 'promo'))
		if commonCrits:
			allCommonCrits += commonCrits
		if name is None:
			name = set_.name
		if structure is None:
			key = set_.packStructure
			if key is not None:
				try:
					structure = NamedPackStructure[key].structure
				except KeyError:
					raise KeyError(f"Tried to creat a pack using '{key}' structure, but it hasn't been defined.")
			else:
				try:
					structure = NamedPackStructure[set_.code].structure
				except KeyError:
					structure = NamedPackStructure['default'].structure
			if set_.code == 'tsp':
				#Time Spiral needs to lose its set-code common crit.
				self.commonCrits = [CommonCrit('Type', 'DOESNTCONTAIN', 'Token'), CommonCrit('DraftQualities', 'DOESNTCONTAIN', 'notInPacks')]
		super().__init__(name=f'{name} Booster', packKind='Booster', structure=structure, commonCrits=allCommonCrits)
	

class TspBooster(Booster):
		#Just for Time Spiral because it's weird and includes two sets.
		def __init__(self, structure):
			super().__init__('Time Spiral', tsp, structure)
			self.commonCrits = [CommonCrit('Type', 'DOESNTCONTAIN', 'Token')]
		

class TitlePack(Pack):
	#An empty pack declared just to give section headings in the list of packs.
	def __init__(self, name):
		super().__init__(name=f'### {name} ###', packKind='Title', structure=[])
	

class SeededBooster(Booster):
	def __init__(self, name, packCode, structure=None):
		set_= SeededPackRecord.getSetCodeForPack(packCode)
		commonCrits = (CommonCrit('DraftQualities', 'CONTAINS', packCode),)
		super().__init__(set_, structure=structure, name=name, commonCrits=commonCrits)
		self.packKind = 'Seeded Prerelease Pack'

class SeededPackRecord(metaclass=Recorded):
	"""Records which cards are in which given seeded packs. Indexed to a set."""
	key = 'set_'
	
	def __init__(self, set_, packLists):
		self.set_ = set_
		self.packLists = packLists
	
	@classmethod
	def getPackNamesForCard(cls, card):
		"""Return a list (even if it's just one, or empty)"""
		try:
			record = cls[card.set_.code]
		except KeyError:
			return []
		
		packNames = []
		for packName, packContents in record.packLists.items():
			if card.name in packContents:
				packNames.append(packName)
		
		return packNames
	
	@classmethod
	def getSetCodeForPack(cls, packCode):
		"""Return the card set corresponding to the specified pack code."""
		#GRUMBLR GRUMBLE GRUMBLE
		from classes import Set
		match = False
		for record in cls.instancesList():
			try:
				record.packLists[packCode]
				match = True
				break
			except KeyError:
				pass
		
		if match:
			return Set[record.set_]
		else:
			raise KeyError(f'No seeded pack at code {packCode}')

if __name__ == '__main__':
	loadPacksConfig()		
