from PIL import Image, ImageDraw, ImageFont, ImageFilter, ImageChops
from helpers import *
import shutil
from io import BytesIO
import math
import re

def splitTextByMana(text):
	line_elements = []
	current_chars = [] #buffer
	for char in text:
		if char == '{':
			if current_chars:
				line_elements.append(''.join(current_chars))
			current_chars = ['{']
		elif char == '}':
			current_chars.append('}')
			line_elements.append(''.join(current_chars))
			current_chars = []
		else:
			current_chars.append(char)
	if current_chars:
		line_elements.append(''.join(current_chars))
	
	return line_elements

def renderSymbols(symbolString, size=19, paddingTop=0):
	"""Return a PIL image of a mana cost or other card symbol(s), scaled to the supplied height."""
	symbols = splitTextByMana(symbolString)
	width = size
	height = size + int(paddingTop)
	scaledPadding = round(128 / size * paddingTop)
	
	img = Image.new('RGBA', (128*len(symbols), 128 + scaledPadding), (0, 0, 0, 0))
	
	x = 0
	for symbol in symbols:
		code = symbol.replace('{', '').replace('}', '').lower()
		symbol_image = Image.open(f'../templates/tokens/mana/mana{code}.png')
		#mask = Image.new('L', symbol_image.size, 200)
		img.paste(symbol_image, (x, 0 + scaledPadding))
		x += 128
	
	img = img.resize((width * len(symbols), height), Image.LANCZOS)
	
	return img

def renderMultipleTextLines(text, font, maxWidth=None, maxHeight=None, color=(0, 0, 0)):
	"""
	Splits text at newlines and calls renderTextLine on each,
	then returns an image of them combined with proper spacing
	between hard newlines.
	If no lines were wrapped, or if there are no periods, text
	is centered.
	"""
	if font.size == 0:
		raise ValueError("Can't use fontSize of 0. Did you set a tiny maxHeight and recur too far?")
	
	spacing = round(.4 * font.size)
	hard_lines = text.split('\n')
	images = []
	total_lines = 0
	wrapping_happened = False
	if not '.' in text:
		center = True
	else:
		center = False
	for hard_line in hard_lines:
		image, sub_line_count = wrapTextLine(hard_line, font, maxWidth=maxWidth, color=color, center=center)
		images.append(image)
		if sub_line_count > 1:
			wrapping_happened = True
		total_lines += sub_line_count
		last_sub_line_count = sub_line_count
	
	if not wrapping_happened and len(hard_lines) == 1:
		center = True
	
	if maxWidth:
		total_width = maxWidth
	else:
		total_width = max(image.width for image in images)
	total_height = (total_lines - last_sub_line_count) * font.size + images[-1].height + (len(images) -1) * spacing
	
	if maxHeight is not None and total_height > maxHeight:
		font = font.font_variant(size=font.size-1)
		return renderMultipleTextLines(text, font, maxWidth=maxWidth, maxHeight=maxHeight)
	
	composite = Image.new('RGBA', (total_width, total_height), (0, 0, 0, 0))
	y = 0
	for image in images:
		if center:
			x = round((total_width - image.width) / 2)
		else:
			x = 0
		composite.paste(image, (x, y))
		y += image.height + spacing
	
	return composite

def wrapTextLine(text, font, maxWidth=None, color=(0, 0, 0), center=True):
	"""
	Return a tuple containing:
		- a PIL Image of a line of text, with mana symbols inserted, and wrapped if exceeding maxWidth
		- an integer denoting how many actual lines the image contains (1 if no wrapping was needed)
	"""
	if '\n' in text:
		raise ValueError('wrapTextLine() cannot render text with newlines in it.')
	
	first_attempt = renderTextLine(text, font)
	if maxWidth is None or maxWidth >= first_attempt.width:
		return (first_attempt, 1)
	
	lines = []
	words_that_fit = []
	last_image_that_fit = None
	first_line = True
	
	for word in text.split(' '):
		attempt = words_that_fit + [word]
		attempt_image = renderTextLine(' '.join(attempt), font, color=color, italicizeAbilityWord=first_line)
		if attempt_image.width > maxWidth:
			if last_image_that_fit is None:
				raise ValueError(f"Impossible for '{word}' to fit on line.")
			lines.append(last_image_that_fit)
			words_that_fit = [word]
			first_line = False #We're not longer starting at the beginning of the supplied line.
		else:
			words_that_fit = attempt
			last_image_that_fit = attempt_image
	#Capture the last line:
	last_line_image = renderTextLine(' '.join(words_that_fit), font)
	lines.append(last_line_image)
			
	
	total_width = max(line.width for line in lines)
	total_height = (len(lines) - 1) * font.size + lines[-1].height #Makes sure we catch any descenders on the last line.
	composite = Image.new('RGBA', (total_width, total_height), (0, 0, 0, 0))
	
	y = 0
	for line in lines:
		if center:
			x = round( (total_width - line.width) / 2 )
		else:
			x = 0
		composite.paste(line, (x, y))
		y += font.size
	
	return (composite, len(lines))

def fitTextLine(text, font, maxWidth=None, align='left', color=(0, 0, 0)):
	"""
	Return a PIL image of a line of text, scaled down to fit maxWidth.
	If text is scaled down, it's aligned to the same bottom edge as the
	full size requested would have been.
	'Center' alignment is only respected if maxWidth is set, so that
	there's a width in which to center it.
	"""
	if '\n' in text:
		raise ValueError('fitTextLine() cannot render text with newlines in it.')
	text_image = renderTextLine(text, font, color=color)
	
	if maxWidth is not None:
		if text_image.width > maxWidth:
			original_height = text_image.height
			new_height = round(original_height * maxWidth / text_image.width)
			text_image = text_image.resize((maxWidth, new_height), Image.LANCZOS)
			container = Image.new('RGBA', (maxWidth, original_height), (*color, 0))
			container.paste(text_image, (0, original_height - new_height))
			text_image = container
		elif align == 'center':
			container = Image.new('RGBA', (maxWidth, text_image.height), (*color, 0))
			x = round((maxWidth - text_image.width) / 2)
			container.paste(text_image, (x, 0))
			text_image = container
	
	return text_image

def renderTextLine(text, font, color=(0, 0, 0), italicizeAbilityWord=False):
	"""Return a PIL image of a single line of text, with mana symbols rendered as needed."""
	if '\n' in text:
		raise ValueError('renderTextLine() cannot render text with newlines in it.')
	
	text = text.replace("'", "’")
	mana_size = round(.9 * font.size)
	mana_height_offset = 0 * font.size #Doesn't have to be an int! More accurate this way
	images = []
	elements = splitTextByMana(text)
	italics = False
	if italicizeAbilityWord:
		ability_match =  re.match(r'\w+? — ', elements[0])
		if ability_match:
			ability_word = ability_match.group()
			elements = [ability_word] + [elements[0].replace(ability_word, '')] + elements[1:]
			italics = True
	
	for element in elements:
		if '{' in element:
			image = renderSymbols(element, mana_size, paddingTop=mana_height_offset)
			images.append(image)
		else:
			y = 0
			if italics:
				textFont = ImageFont.truetype('../templates/tokens/fonts/mplantin-italic.ttf', font.size)
				italics = False
			else:
				textFont = font
			
			if font.path == '../templates/tokens/fonts/beleren.ttf':
				image = customBelerenGlyphs(element, font=textFont, color=color)
			else:
				image = Image.new('RGBA', textFont.getsize(element), (*color, 0))
				draw = ImageDraw.Draw(image)
				draw.text((0, y), element, font=textFont, fill=color)
			images.append(image)
	
	return compositeTextImages(images)

def compositeTextImages(images):
	total_height = max(image.height for image in images)
	total_width = sum(image.width for image in images)
	composite = Image.new('RGBA', (total_width, total_height), (0, 0, 0, 0))
	x = 0
	for image in images:
		composite.paste(image, (x, 0))
		x += image.width
	
	return composite

def customBelerenGlyphs(text, font, color):
	elements = re.split(r'([fhmn](?!\w)|\*)', text)
	images = []
	
	for element in filter(None, elements):
		if element in ['f', 'h', 'm', 'n', '*']:
			glyph = Image.open(f'../templates/tokens/fonts/beleren-glyphs/{element}.png').convert('L')
			mask = ImageChops.invert(glyph)
			image = Image.new('RGBA', mask.size, (*color, 255))
			applyMask(image, mask)
			if font.size != 200:
				ratio = font.size / 200
				new_width = round(image.width * ratio)
				new_height = round(image.height * ratio)
				image = image.resize((new_width, new_height), Image.LANCZOS)
		else:
			image = Image.new('RGBA', font.getsize(element), (*color, 0))
			draw = ImageDraw.Draw(image)
			draw.text((0, 0), element, font=font, fill=color)
		images.append(image)
	
	return compositeTextImages(images)

def applyMask(image, mask):
	"""
	Takes an image (with alpha channel) and a greyscale mask.
	Applies the mask to the image's alpha channel.
	"""
	current_alpha = image.getchannel('A')
	composite_alpha = ImageChops.darker(current_alpha, mask)
	image.putalpha(composite_alpha)

class TokenImage():
	"""An image for a custom Lackey token"""
	
	def __init__(self, card, type_line_shift=None):
		self.card = card
		self.artist = None
		self.custom_image_file = None
		
		self.art_layer = None
		self.frame_layers = []
		self.text_layers = []
		
		self.template_size = (745, 1040)
		
		display(3, f"Creating token image for '{card.lackeyName}'")
		
		#Get a custom token image, if I've made one:
		customFiles = os.listdir('../templates/tokens/custom/')
		fileName = card.tokenImageFileName(includeFolder=False, includeVersion=False) + '.jpg'
		if fileName in customFiles:
			display(3, 'Using custom presaved image.')
			self.frame_layers.append(Image.open('../templates/tokens/custom/' + fileName).convert('RGBA'))
			return
		
		self.setFrame()
		
		self.addBackground()
		
		self.addFramePart('title_line', frame='other')
		
		self.addTitleLineText()
		
		if card.text:
			self.type_line_shift = 118 #Normal token placement; 0 is for an actual card
			if type_line_shift is not None:
				#Custom setting, mostly just for generating the Jace overlay.
				self.type_line_shift = type_line_shift
			self.addCardText() #shifts type line up if necessary
			self.addTextBox()
		else:
			self.type_line_shift = 265 #"Full" art for vanilla
		
		self.addTypeLine()
		
		self.addTypeLineText()
		
		self.addFramePart('set_symbol', frame='other', y_offset=self.type_line_shift)
		
		self.addArt()
		
		if self.card.isType('creature'):
			self.addPowerToughnessBox()
			ptFont = ImageFont.truetype('../templates/tokens/fonts/beleren.ttf', 41)
			pt = '/'.join((self.card.power, self.card.toughness))
			self.addLine((593, 930), pt, font=ptFont, align='center', maxWidth=95)
		
		# bottomFont = ImageFont.truetype('/Library/Application Support/iLifeSlideshow/v2/Fonts/BlairMdITC TT-Medium.dfont', 18)
		# 		whitish = (190,) * 3
		# 		
		# 		if self.artist:
		# 			self.addFramePart('paintbrush', frame='other')
		# 			self.addLine((80, 971), self.artist, font=bottomFont, color=whitish)
		# 		
		# 		self.addLine((49, 994), 'Custom  Token  for  LackeyCCG', font=bottomFont, color=whitish)
		
		beleren_caps = ImageFont.truetype('../templates/tokens/fonts/beleren-small-caps.ttf', 23)
		mplantin = ImageFont.truetype('../templates/tokens/fonts/mplantin-bold.ttf', 30)
		whitish = (210,) * 3
		
		if self.artist:
			self.addFramePart('paintbrush', frame='other')
			self.addLine((80, 968), self.artist, font=beleren_caps, color=whitish)
		
		self.addLine((49, 988), 'Custom LackeyCCG token, from content © Wizards of the Coast', font=mplantin, color=whitish, maxWidth=645)
	
	def __repr__(self):
		return f"<TokenImage for '{self.card.lackeyName}'>"
	
	def addLine(self, position, text, font, maxWidth=None, align='left', color=(0, 0, 0)):
		"""
		Adds a new text layer with a line of text at position (x, y).
		"""
		text_image = fitTextLine(text, font, maxWidth=maxWidth, align=align, color=color)
		text_overlay = Image.new('RGBA', self.template_size, (*color, 0))
		text_overlay.paste(text_image, position)
		self.text_layers.append(text_overlay)
	
	def addTitleLineText(self):
		"""Write the name, and the mana cost if there is one."""
		name_width = 610
		name_align = 'center'
		
		if self.card.cost != '':
			mana = renderSymbols(self.card.cost, 38)
			overlay = Image.new('RGBA', self.template_size, (0, 0, 0, 0))
			x = 685 - mana.width
			overlay.paste(mana, (x, 61))
			self.text_layers.append(overlay)
			
			name_width -= mana.width
			name_align = 'left'
		
		font = ImageFont.truetype('../templates/tokens/fonts/beleren-small-caps.ttf', 43)
		self.addLine((65, 52), self.card.name, font=font, align=name_align, maxWidth=name_width, color=(242, 227, 111))
	
	def setFrame(self):
		card = self.card
		
		colors = {
			'': 'colorless',
			'W': 'white',
			'U': 'blue',
			'B': 'black',
			'R': 'red',
			'G': 'green',
		}
		
		color = card.color
		self.second_frame = None
		
		self.frame = colors.get(color, None)	
		if self.frame is None:	
			if len(color) == 2:
				self.frame = colors[color[0]]
				self.second_frame = colors[color[1]]
			elif len(color) > 2:
				self.frame = 'gold'
		
		if card.isType('artifact'):
			if self.frame == 'colorless':
				self.frame = 'artifact'
	
	def addFramePart(self, frame_part, frame=None, y_offset=0, mask_top=0, mask=None, addAsMask=False):
		if frame is None:
			frame = self.frame
			default_frame = True
		else:
			default_frame = False
		frame_image = Image.open(f'../templates/tokens/frames/{frame}/{frame_part}.png')
		
		if default_frame and self.second_frame:
			if frame_part in ['background', 'type_line', 'text_box']:
				right_half = Image.open(f'../templates/tokens/frames/{self.second_frame}/{frame_part}.png')
				hybrid_mask = Image.open(f'../templates/tokens/frames/other/multicolor_mask.png').convert("L")
				frame_image.paste(right_half, hybrid_mask)
		
		if mask is not None:
			applyMask(frame_image, mask)
		
		if y_offset != 0:
			blank = Image.new('RGBA', self.template_size, (0, 0, 0, 0))
			blank.paste(frame_image, (0, y_offset))
			frame_image = blank
		
		if mask_top != 0:
			mask = Image.new('L', self.template_size, 255)
			draw = ImageDraw.Draw(mask)
			draw.rectangle([(0, 0), (frame_image.width, mask_top)], fill=0)
			
			blank = Image.new('RGBA', self.template_size, (0, 0, 0, 0))
			blank.paste(frame_image, (0, 0), mask=mask)
			frame_image = blank
		
		if addAsMask:
			frame_image = ImageChops.invert(frame_image.getchannel('A'))
		
		self.frame_layers.append(frame_image)
	
	def addBackground(self):
		#Add code here to apply artifact background over pinstriping
		if self.card.isType('enchantment creature'):
			self.addFramePart('background_enchantment')
		else:
			self.addFramePart('background')
		if self.card.isType('artifact') and self.frame != 'artifact':
			artifact_mask = Image.open('../templates/tokens/frames/artifact/colored_artifact_mask.png').convert('L')
			self.addFramePart('background', frame='artifact', mask=artifact_mask)
	
	def addTextBox(self):
		if self.frame == 'colorless':
			self.addFramePart('text_box', frame='artifact', mask_top=642+self.type_line_shift, addAsMask=True)
		self.addFramePart('text_box', mask_top=642+self.type_line_shift)
	
	def addTypeLine(self):
		type_line_mask = None
		if self.card.text != '':
			type_line_mask = Image.open('../templates/tokens/frames/other/type_line_mask.png').convert('L')
		
		if self.frame == 'colorless':
			self.addFramePart('type_line', frame='artifact', y_offset=self.type_line_shift, mask=type_line_mask, addAsMask=True)
		
		self.addFramePart('type_line', y_offset=self.type_line_shift, mask=type_line_mask)
		
		if self.card.isType('artifact') and (self.frame == 'gold' or self.second_frame is not None):
			gold_mask = Image.open(f'../templates/tokens/frames/hybrid/type_line.png').getchannel('A')
			self.addFramePart('type_line', frame='gold', y_offset=self.type_line_shift, mask=gold_mask)
		elif self.second_frame is not None:
			self.addFramePart('type_line', frame='hybrid', y_offset=self.type_line_shift)
	
	def addPowerToughnessBox(self):
		#Code here to deal with artifact color
		frame = None
		if self.second_frame is not None:
			if self.card.isType('artifact'):
				frame = 'gold'
			else:
				frame = 'hybrid'
		self.addFramePart('power_toughness', frame=frame)
	
	@indented
	def addArt(self):
		if self.frame == 'colorless':
			y = 26
			targetHeight = 945
			targetWidth = 692
		else:
			y = 125
			targetHeight = 455 + self.type_line_shift
			targetWidth = 635
		
		targetAspect = targetWidth / targetHeight
		#display(3, f'Target aspect ratio: {round(targetAspect, 2)}')
		
		art = None
		self.artist = None
		artDir = '../templates/tokens/art/'
		isCounter = self.card.isType('counter')
		if isCounter:
			artDir += 'counters/'
			fileName = self.card.name.replace(' Counter', '').replace(' ', '-').lower()
		else:
			fileName = self.card.tokenImageFileName(includeFolder=False, includeVersion=False)
		artFiles = os.listdir(artDir)
		for artFile in artFiles:
			if fileName in artFile:
				art = Image.open(artDir + artFile)
				break
		if art is None and self.card.isCopy:
			card = self.card
		
			params = {
				'q': f'!"{card.name}" is:hires',
				'unique': 'cards',
			}
			request = scryfallRequest(params)
			if request:
				jsonCard = request[0]
			else:
				params['q'] = f'!"{card.name}"'
				jsonCard = scryfallRequest(params)[0]
		
			display(3, f'Downloading art image for token copy of {card.name}')
			response = requests.get(jsonCard['image_uris']['art_crop'])
			art = Image.open(BytesIO(response.content))
			self.artist = jsonCard['artist']
			artFile = f'../templates/tokens/art/{fileName} by {self.artist}.jpg'
			art.save(artFile)
		
		if art:
			width, height = art.size
			aspect = width / height
			#display(3, f'Aspect of art file: {round(aspect, 2)}')
			
			if aspect >= targetAspect:
				#display(3, 'Art is wider. Fit to height.')
				conversion = targetHeight / height
			else:
				#display(3, 'Art is taller. Fit to width.')
				conversion = targetWidth / width
			
			template_width, template_height = self.template_size
			new_width, new_height = round(template_width / conversion), round(template_height / conversion)
		
			#art = art.resize(newSize, Image.LANCZOS)
		
			self.art_layer = Image.new('RGBA', (new_width, new_height), (255, 255, 255, 0))
			position = (round( (new_width - art.width) / 2 ), round(y / conversion))
			self.art_layer.paste(art, position)
			
			#Setting artist's name
			if isCounter:
				self.artist = 'Background Texture by Hillary Luetkemeyer'
			elif self.artist is None:
				if not f"{fileName} by " in artFile:
					raise NameError(f"Art file for {self.card.lackeyName} doesn't have properly formatted artist credit.")
				self.artist = artFile.replace(f'{fileName} by ', '')[:-4]
	
	def addCardText(self, fontSize=38):
		box_top = 784
		box_width = 603
		margin = 64
		box_height = 162
		minimum_size = 34
		#Remove this:
		#minimum_size = 100
		pt_box = (564, 905, 681, 966)
		
		#print(f'Font size: {fontSize}')
		font = ImageFont.truetype('../templates/tokens/fonts/mplantin.ttf', fontSize)
		if re.match(r'\(.+\)', self.card.text):
			font = ImageFont.truetype('../templates/tokens/fonts/mplantin-italic.ttf', fontSize)
		text = re.sub(r'"(.+?)"', r'“\1”', self.card.text)
		text_image = renderMultipleTextLines(text, font, maxWidth=box_width)
		#text_image.getchannel('A').show()
		if text_image.height > box_height and fontSize > minimum_size:
			#print(f'Going to try a font size: {fontSize - 1}')
			self.addCardText(fontSize=fontSize-1)
			return
		if text_image.height > box_height + self.type_line_shift:
			#print('Making text as big as will fit in full-height box.')
			text_image = renderMultipleTextLines(text, font, maxWidth=box_width, maxHeight=box_height+self.type_line_shift)
			change = text_image.height - box_height
			box_height += change
			box_top -= change
			#print(f'Changing type line shift to {self.type_line_shift}')
			self.type_line_shift -= change
		elif text_image.height > box_height:
			excess = text_image.height - box_height
			self.type_line_shift -= excess
			#print(f'Shifting type line up by {excess} pixels.')
			box_height = text_image.height
			box_top -= excess
		
		text_overlay = Image.new('RGBA', self.template_size, (0, 0, 0, 0))
		y = round((box_height - text_image.height) / 2 + box_top)
		#draw = ImageDraw.Draw(self.img)
		#draw.rectangle([(margin, box_top), (margin + box_width, box_top + box_height)], fill=(0, 200, 0, 100))
		text_overlay.paste(text_image, (margin, y))
		
		if self.card.isType('creature'):
			#Check for P/T overlap.
			danger_zone = text_overlay.crop(pt_box).getchannel('A')
			interference = danger_zone.getbbox()
			if interference is not None:
				interference = interference[3] #Vertical component
				#print(f'Overlap with P/T box by {interference} pixels detected.')
				if fontSize > minimum_size or self.type_line_shift - interference <= 0:
					self.addCardText(fontSize=fontSize-2)
					return
				else:
					self.type_line_shift -= interference
					#print(f'Shifting type line up by {interference} more pixels ({self.type_line_shift} total).')
					#Leave box_height the same, so the text shifts up.
					text_overlay = Image.new('RGBA', self.template_size, (0, 0, 0, 0))
					text_overlay.paste(text_image, (margin, y - interference))
					
		self.text_layers.append(text_overlay)
	
	def addTypeLineText(self, y=None, maxWidth=567):
		font = ImageFont.truetype('../templates/tokens/fonts/beleren.ttf', 34)
		if y is None:
			y = 595 + self.type_line_shift
		type_ = self.card.type_
		if 'Counter' in type_:
			type_ = 'Counter'
		self.addLine((62, y), type_, font=font, maxWidth=maxWidth)
	
	@indented
	def save(self, size=(312, 445), crop=True, fileName=None, forcepng=False, fullsize=False):
		if fullsize:
			size = self.template_size
			crop = False
		if hasattr(self, 'type_line_shift') and self.type_line_shift > 118 and self.type_line_shift < 265:
			print('TYPE LINE IS IN THE DANGER ZONE')
		if fileName is None:
			fileName = os.path.join('../images', self.card.tokenImageFileName() + '.jpg')
		if forcepng:
			fileName = fileName.replace('.jpg', '.png')
		
		os.makedirs(os.path.dirname(fileName), exist_ok=True)
		display(3, f'Saving image for {self.card} as {fileName}')
		if self.art_layer:
			art = self.art_layer
		else:
			art = Image.new('RGBA', self.template_size, (0, 0, 0, 0))
		
		frame_and_text = Image.new('RGBA', self.template_size, (0, 0, 0, 0))
		for layer in self.frame_layers + self.text_layers:
			if layer.mode == 'L':
				applyMask(frame_and_text, layer)
			else:
				frame_and_text.alpha_composite(layer)
		
		if crop:
			fullsize_crop = (18, 14, 727, 1024)
			conversion = art.width / frame_and_text.width
			art_crop = tuple(round(dimension * conversion) for dimension in fullsize_crop)
			frame_and_text = frame_and_text.crop(fullsize_crop)
			art = art.crop(art_crop)
		
		art = art.resize(size, Image.LANCZOS)
		#frame_and_text.show()
		frame_and_text = frame_and_text.resize(size, Image.LANCZOS)
		#mask = Image.new('L', size, 50)
		#applyMask(frame_and_text, mask)
		
		image = Image.alpha_composite(art, frame_and_text)
		
		if fileName[-3:] == 'jpg':
			image = image.convert('RGB')
			image.save(fileName, quality=85)
		else:
			image.save(fileName)
